from keopscore.formulas.VectorizedComplexScalarOp import VectorizedComplexScalarOp
from keopscore.formulas.complex.Conj import Conj

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexMult                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexMult(VectorizedComplexScalarOp):

    string_id = "ComplexMult"

    def ScalarOp(self, out, inF, inG):
        string = out[0].assign(inF[0] * inG[0] - inF[1] * inG[1])
        string += out[1].assign(inF[0] * inG[1] + inF[1] * inG[0])
        return string

    def DiffT(self, v, gradin):
        f, g = self.children
        return f.DiffT(v, ComplexMult(Conj(g), gradin)) + g.DiffT(
            v, ComplexMult(Conj(f), gradin)
        )



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_powf


class Powf(VectorizedScalarOp):
    """the Power vectorized operation"""

    string_id = "Powf"

    ScalarOpFun = keops_powf

    @staticmethod
    def Derivative(a, b):
        from keopscore.formulas.maths.Log import Log

        return b * Powf(a, b - 1), Log(a) * Powf(a, b)

    # parameters for testing the operation (optional)
    test_ranges = [(0, 2), (-1, 1)]  # ranges of arguments
    torch_op = "lambda x,y : torch.pow(x, y)"



from keopscore.formulas.Operation import Operation
from keopscore.formulas.maths.Extract import Extract
from keopscore.utils.code_gen_utils import c_variable, c_for_loop, c_zero_float
from keopscore.utils.code_gen_utils import c_array, VectCopy
from keopscore.utils.misc_utils import KeOps_Error

# //////////////////////////////////////////////////////////////
# ////       BSPLINE VECTOR : BSPLINE<knots,x,order>        ////
# //////////////////////////////////////////////////////////////

# N.B.: BSpline_Impl creates a vector of size "n_knots - 1"
# since we need this amount of space as an intermediate buffer for the computation
# of the final "n_knots - order - 1" coefficients.
# In the final BSpline operation, we simply use "extract" to discard
# the "order" irrelevant coefficients.


class BSpline_Impl(Operation):
    string_id = "BSpline_Impl"

    def __init__(self, knots, x, order):
        if order < 0:
            KeOps_Error(f"The order of a BSpline must be >= 0, but received {order}.")

        if x.dim != 1:
            KeOps_Error(
                f"BSplines must be sampled at scalar values, "
                f"but received a vector of size {x.dim}."
            )

        if knots.dim < order + 2:
            KeOps_Error(
                f"BSplines of order {order} require at least {order+2} knots, "
                f"but only received a vector of size {knots.dim}."
            )

        super().__init__(knots, x, params=(order,))
        self.order = order
        self.dim = knots.dim - 1

    def Op(self, out, table, knots, inX):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        x = c_variable("float")
        ratio_1 = c_variable("float")
        ratio_2 = c_variable("float")
        init_loop, j = c_for_loop(0, self.dim, 1, pragma_unroll=True)
        outer_loop, k = c_for_loop(1, self.order + 1, 1, pragma_unroll=True)
        inner_loop, i = c_for_loop(
            0, c_variable("int", self.dim) - k, 1, pragma_unroll=True
        )

        inner_loop_code = f"""
        // Compute the second ratio "omega_i+1,k" = (x - t_i+1) / (t_i+k+1 - t_i+1):
        {ratio_2.assign( 
            (knots[i+1] < knots[i+1+k]).ternary(
            (x - knots[i+1]) / (knots[i+1+k] - knots[i+1]),
            c_zero_float,)
        )}
        // In place computation of B_i,k+1(x) as
        // omega_i,k(x) * B_i,k(x) + (1 - omega_i+1,k(x)) * B_i+1,k(x)
        {out[i].assign(
            ratio_1 * out[i] + (c_variable("float", "1.0f") - ratio_2) * out[i+1])}

        // Update the ratios as i -> i+1:
        {ratio_1.assign(ratio_2)}
        """

        code = f"""
            // The sampling 'time':
            {x.declare_assign(inX[0])}
            // Order 1 Spline: one-hot encoding of whether t[j] <= x < t[j+1]
            {init_loop(out[j].assign((knots[j] <= x).logical_and(x < knots[j+1])))}

            // Recursive De Boor's algorithm:
            {ratio_1.declare()}
            {ratio_2.declare()}
            {outer_loop(
                f'''// Compute the first ratio "omega_i,k" = (x - t_i) / (t_i+k - t_i) for i=0:
                {ratio_1.assign( 
                    (knots[0] < knots[k]).ternary( 
                    (x - knots[0]) / (knots[k] - knots[0]),
                    c_zero_float, ) 
                    )} 

                // Loop over out[0:len(out)-k]
                {inner_loop(inner_loop_code)}
        ''')}
        """
        return code

    def DiffT(self, v, gradin):
        raise NotImplementedError(
            "KeOps BSplines are not yet differentiable."
            "Please consider disabling autograd for the relevant variables"
            "with 'x.detach()' "
            "or submit a feature request at https://github.com/getkeops/keops/issues."
        )

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 2  # number of arguments
    test_argdims = [5, 1]  # dimensions of arguments for testing
    test_params = [3]  # values of parameters for testing
    torch_op = None  # equivalent PyTorch operation


# TODO Jean: The two ways of defining an alias seem equivalent to me...
#            Which one should we use?
if True:

    def BSpline(knots, x, order):
        return Extract(BSpline_Impl(knots, x, order), 0, knots.dim - order - 1)

else:

    class BSpline:
        def __new__(cls, knots, x, order):
            return Extract(BSpline_Impl(knots, x, order), 0, knots.dim - order - 1)

        enable_test = False



from keopscore.utils.code_gen_utils import VectCopy
from keopscore.formulas.reductions.Min_ArgMin_Reduction_Base import (
    Min_ArgMin_Reduction_Base,
)
from keopscore.formulas.reductions.Zero_Reduction import Zero_Reduction


class ArgMin_Reduction(Min_ArgMin_Reduction_Base):
    """Implements the argmin reduction operation : for each i or each j, find the index of the
    minimal value of Fij operation is vectorized: if Fij is vector-valued, argmin is computed for each dimension."""

    string_id = "ArgMin_Reduction"

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)
        self.dim = formula.dim

    def FinalizeOutput(self, acc, out, i):
        acc_val, acc_ind = acc.split(self.dim, self.dim)
        return VectCopy(out, acc_ind)

    def DiffT(self, v, gradin):
        return Zero_Reduction(v.dim, v.cat % 2)



from keopscore.binders.LinkCompile import LinkCompile


class Cpu_link_compile(LinkCompile):

    source_code_extension = "cpp"

    def __init__(self):
        LinkCompile.__init__(self)
        # these are used for command line compiling mode
        self.low_level_code_file = "".encode("utf-8")

        # actual dll to be called
        self.true_dllname = self.gencode_file
        # file to check for existence to detect compilation is needed
        self.file_to_check = self.gencode_file

    def generate_code(self):
        # method to generate the code and compile it
        # generate the code and save it in self.code, by calling get_code method from CpuReduc class :
        self.get_code()
        # write the code in the source file
        self.write_code()
        # retreive some parameters that will be saved into info_file.
        self.tagI = self.red_formula.tagI
        self.dim = self.red_formula.dim



from keopscore.formulas.Operation import Operation
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.code_gen_utils import value
from keopscore.utils.misc_utils import KeOps_Error

##########################
######    SumT       #####
##########################


class SumT_Impl(Operation):
    # the adjoint of the summation operation

    string_id = "SumT"

    def __init__(self, arg, dim):
        super().__init__(arg, params=(dim,))
        self.dim = dim

    def Op(self, out, table, arg):
        return out.assign(value(arg))

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.Sum import Sum

        f = self.children[0]
        return f.DiffT(v, Sum(gradin))


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def SumT(arg, dim):
    if arg.dim != 1:
        KeOps_Error("dimension of argument must be 1 for SumT operation")
    elif isinstance(arg, Zero):
        return Zero(dim)
    else:
        return SumT_Impl(arg, dim)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_ifelse


class IfElse(VectorizedScalarOp):
    """the if/else vectorized operation
    IfElse(f,a,b) = a if f>=0, b otherwise
    """

    string_id = "IfElse"

    ScalarOpFun = keops_ifelse

    def DiffT(self, v, gradin):
        f, a, b = self.children
        return IfElse(f, a.DiffT(v, gradin), b.DiffT(v, gradin))

    # parameters for testing the operation (optional)
    nargs = 3  # number of arguments



from keopscore import debug_ops_at_exec
from keopscore.binders.cpp.Cpu_link_compile import Cpu_link_compile

from keopscore.mapreduce.cpu.CpuAssignZero import CpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_array,
    c_include,
)
import keopscore


class CpuReduc_ranges(MapReduce, Cpu_link_compile):
    # class for generating the final C++ code, Cpu version

    AssignZero = CpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Cpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        i = self.i
        j = self.j
        dtype = self.dtype
        red_formula = self.red_formula
        fout = self.fout
        outi = self.outi
        acc = self.acc
        acctmp = self.acctmp
        arg = self.arg
        args = self.args
        nargs = len(args)

        xi = self.xi
        yj = c_array(dtype, self.varloader.dimy, "yj")
        param_loc = self.param_loc

        varloader = self.varloader
        table = varloader.table(xi, yj, param_loc)

        nvarsi, nvarsj, nvarsp = (
            len(self.varloader.Varsi),
            len(self.varloader.Varsj),
            len(self.varloader.Varsp),
        )

        tagHostDevice, tagCpuGpu, tag1D2D = (
            self.tagHostDevice,
            self.tagCpuGpu,
            self.tag1D2D,
        )

        sum_scheme = self.sum_scheme

        indices_i = c_array("int", nvarsi, "indices_i")
        indices_j = c_array("int", nvarsj, "indices_j")
        indices_p = c_array("int", nvarsp, "indices_p")
        imstartx = c_variable("int", "i-start_x")
        jmstarty = c_variable("int", "j-start_y")

        headers = ["cmath", "stdlib.h"]
        if keopscore.config.config.use_OpenMP:
            headers.append("omp.h")
        if debug_ops_at_exec:
            headers.append("iostream")
        self.headers += c_include(*headers)

        self.code = f"""
{self.headers}

#define do_keops_checks 0
#if do_keops_checks
    #include <string>
    #include <iostream>
#endif

#if do_keops_checks
    #define Error_msg_no_cuda "[KeOps] This KeOps shared object has been compiled without cuda support: - 1) to perform computations on CPU, simply set tagHostDevice to 0 - 2) to perform computations on GPU, please recompile the formula with a working version of cuda."
    void keops_error(std::string message) {{
    	throw std::runtime_error(message);
    }}
#endif

#if do_keops_checks
void check_nargs(int nargs, int nminargs) {{
    if(nargs<nminargs) {{
        keops_error("[KeOps] : not enough input arguments");
    }}
}}
#endif

#include "include/Sizes.h"
#include "include/ranges_utils.h"
#include "include/Ranges.h"

template< typename TYPE>                 
int CpuConv_ranges_{self.gencode_filename}(int nx, int ny, 
                    int nbatchdims, int* shapes,
                    std::vector< int > indsi, std::vector< int > indsj, std::vector< int > indsp,
                    int nranges_x, int nranges_y, int **ranges,
                    TYPE* out, TYPE **{arg.id}) {{
                        
    int sizei = indsi.size();
    int sizej = indsj.size();
    int sizep = indsp.size();
    
    // Separate and store the shapes of the "i" and "j" variables + parameters --------------
    //
    // shapes is an array of size (1+nargs)*(nbatchdims+3), which looks like:
    // [ A, .., B, M, N, D_out]  -> output
    // [ A, .., B, M, 1, D_1  ]  -> "i" variable
    // [ A, .., B, 1, N, D_2  ]  -> "j" variable
    // [ A, .., B, 1, 1, D_3  ]  -> "parameter"
    // [ A, .., 1, M, 1, D_4  ]  -> N.B.: we support broadcasting on the batch dimensions!
    // [ 1, .., 1, M, 1, D_5  ]  ->      (we'll just ask users to fill in the shapes with *explicit* ones)
            
    int shapes_i[sizei * (nbatchdims + 1)], shapes_j[sizej * (nbatchdims + 1)], shapes_p[sizep * (nbatchdims + 1)];

    
    // First, we fill shapes_i with the "relevant" shapes of the "i" variables,
    // making it look like, say:
    // [ A, .., B, M]
    // [ A, .., 1, M]
    // [ A, .., A, M]
    // Then, we do the same for shapes_j, but with "N" instead of "M".
    // And finally for the parameters, with "1" instead of "M".
    fill_shapes(nbatchdims, shapes, shapes_i, shapes_j, shapes_p,  {red_formula.tagJ}, indsi, indsj, indsp);
    
    // Actual for-for loop -----------------------------------------------------

    {param_loc.declare()}
    {varloader.load_vars("p", param_loc, args)}  // If nbatchdims == 0, the parameters are fixed once and for all
    
    // Set the output to zero, as the ranges may not cover the full output -----
    {acctmp.declare()} // __TYPEACC__ acctmp[DIMRED];
    for (int i = 0; i < nx; i++) {{
        {red_formula.InitializeReduction(acctmp)}
        {red_formula.FinalizeOutput(acctmp, outi, i)}
    }}
    
    
    // N.B.: In the following code, we assume that the x-ranges do not overlap.
    //       Otherwise, we'd have to assume that DIMRED == DIMOUT
    //       or allocate a buffer of size nx * DIMRED. This may be done in the future.
    // Cf. reduction.h: 
    //    FUN::tagJ = 1 for a reduction over j, result indexed by i
    //    FUN::tagJ = 0 for a reduction over i, result indexed by j

    int nranges = {red_formula.tagJ} ? nranges_x : nranges_y;
    int* ranges_x = {red_formula.tagJ} ? ranges[0] : ranges[3];
    int* slices_x = {red_formula.tagJ} ? ranges[1] : ranges[4];
    int* ranges_y = {red_formula.tagJ} ? ranges[2] : ranges[5];

    int indices_i[sizei], indices_j[sizej], indices_p[sizep];  // Buffers for the "broadcasted indices"
    for (int k = 0; k < sizei; k++) {{ indices_i[k] = 0; }}  // Fill the "offsets" with zeroes,
    for (int k = 0; k < sizej; k++) {{ indices_j[k] = 0; }}  // the default value when nbatchdims == 0.
    for (int k = 0; k < sizep; k++) {{ indices_p[k] = 0; }}
    
    
    for (int range_index = 0; range_index < nranges; range_index++) {{
        int start_x = ranges_x[2 * range_index];
        int end_x = ranges_x[2 * range_index + 1];
        int start_slice = (range_index < 1) ? 0 : slices_x[range_index - 1];
        int end_slice = slices_x[range_index];

        // If needed, compute the "true" start indices of the range, turning
        // the "abstract" index start_x into an array of actual "pointers/offsets" stored in indices_i:
        if (nbatchdims > 0) {{
            vect_broadcast_index(start_x, nbatchdims, sizei, shapes, shapes_i, indices_i);
            // And for the parameters, too:
            vect_broadcast_index(range_index, nbatchdims, sizep, shapes, shapes_p, indices_p);
            {varloader.load_vars("p", param_loc, args, offsets=indices_p)}  // Load the paramaters, once per tile
        }}
    
        #pragma omp parallel for   
        for (int i = start_x; i < end_x; i++) {{
            {xi.declare()}
            {yj.declare()}
            {fout.declare()}
            {acc.declare()}
            {sum_scheme.declare_temporary_accumulator()}
            if (nbatchdims == 0) {{
                {varloader.load_vars("i", xi, args, row_index=i)}
            }} else {{
                {varloader.load_vars("i", xi, args, row_index=imstartx, offsets=indices_i)}
            }}
            {red_formula.InitializeReduction(acc)}
            {sum_scheme.initialize_temporary_accumulator()}
            for (int slice = start_slice; slice < end_slice; slice++) {{
                 int start_y = ranges_y[2 * slice];
                 int end_y = ranges_y[2 * slice + 1];

                // If needed, compute the "true" start indices of the range, turning
                // the "abstract" index start_y into an array of actual "pointers/offsets" stored in indices_j:
                if (nbatchdims > 0) {{
                    vect_broadcast_index(start_y, nbatchdims, sizej, shapes, shapes_j, indices_j);
                }}
                if (nbatchdims == 0) {{
                    for (int j = start_y; j < end_y; j++) {{
                        {varloader.load_vars("j", yj, args, row_index=j)}
                        {red_formula.formula(fout,table)}
                        {sum_scheme.accumulate_result(acc, fout, j)}
                    }}
                }} else {{
                    for (int j = start_y; j < end_y; j++) {{
                        {varloader.load_vars("j", yj, args, row_index=jmstarty, offsets=indices_j)}
                        {red_formula.formula(fout,table)}
                        {sum_scheme.accumulate_result(acc, fout, jmstarty)}
                    }}
                }}
            }}
            {sum_scheme.final_operation(acc)}
            {red_formula.FinalizeOutput(acc, outi, i)}
        }}
    }}
    return 0;
}}
                    """

        self.code += f"""    
 
#include "stdarg.h"
#include <vector>

template < typename TYPE >
int launch_keops_{self.gencode_filename}(int nx, int ny, 
                                         int tagI, int use_half, int  dimred,
                                         int use_chunk_mode,
                                         std::vector< int > indsi, std::vector< int > indsj, std::vector< int > indsp,
                                         int dimout,
                                         std::vector< int > dimsx, std::vector< int > dimsy, std::vector< int > dimsp,
                                         int **ranges, 
                                         TYPE *out, int nargs, TYPE** arg,
                                         std::vector<std::vector< int >> argshape) {{
    
    Sizes< TYPE > SS (nargs, arg, argshape, nx, ny,tagI, use_half,
                      dimout,
                      indsi, indsj, indsp,
                      dimsx, dimsy, dimsp
                      );
    
    if (use_half)
      SS.switch_to_half2_indexing();

    Ranges < TYPE > RR(SS, ranges);
    
    nx = SS.nx;
    ny = SS.ny;
    
    if (tagI==1) {{
        int tmp = ny;
        ny = nx;
        nx = tmp;
    
        std::vector< int > tmp_v;
        
        tmp_v = indsj;
        indsj = indsi;
        indsi = tmp_v;

        tmp_v = dimsy;
        dimsy = dimsx;
        dimsx = tmp_v;
    }}
    
    return CpuConv_ranges_{self.gencode_filename}< TYPE> (nx, ny, SS.nbatchdims, SS.shapes,
                                                          indsi, indsj, indsp,
                                                          RR.nranges_x, RR.nranges_y, RR.castedranges,
                                                          out, arg);
}}

template < typename TYPE >
int launch_keops_cpu_{self.gencode_filename}(int dimY,
                                             int nx,
                                             int ny,
                                             int tagI,
                                             int tagZero,
                                             int use_half,
                                             int dimred,
                                             int use_chunk_mode,
                                             std::vector< int > indsi, std::vector< int > indsj, std::vector< int > indsp,
                                             int dimout,
                                             std::vector< int > dimsx, std::vector< int > dimsy, std::vector< int > dimsp,
                                             int **ranges,
                                             std::vector< int > shapeout, TYPE *out,
                                             TYPE **arg,
                                             std::vector< std::vector< int > > argshape) {{
    

    
    return launch_keops_{self.gencode_filename}< TYPE >(nx,
                                                        ny,
                                                        tagI,
                                                        use_half,
                                                        dimred,
                                                        use_chunk_mode,
                                                        indsi, indsj, indsp,
                                                        dimout,
                                                        dimsx, dimsy, dimsp,
                                                        ranges,
                                                        out, 
                                                        argshape.size(), 
                                                        arg, 
                                                        argshape);
}}
                        
                """



from keopscore import debug_ops_at_exec
from keopscore.binders.cpp.Cpu_link_compile import Cpu_link_compile
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    c_include,
    c_zero_float,
)
import keopscore


class CpuAssignZero(MapReduce, Cpu_link_compile):
    # class for generating the final C++ code, Cpu version

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Cpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        outi = self.outi
        dtype = self.dtype
        arg = self.arg
        args = self.args

        headers = ["stdlib.h"]
        if keopscore.config.config.use_OpenMP:
            headers.append("omp.h")
        if debug_ops_at_exec:
            headers.append("iostream")
        self.headers += c_include(*headers)

        self.code = f"""
{self.headers}

template < typename TYPE >
int AssignZeroCpu_{self.gencode_filename}(int nx, int ny, TYPE* out,  TYPE **{arg.id}) {{
    #pragma omp parallel for
    for (int i = 0; i < nx; i++) {{
        {outi.assign(c_zero_float)}
    }}
    return 0;
}}

#include "stdarg.h"
#include <vector>

template < typename TYPE >
int launch_keops_{self.gencode_filename}(int nx, int ny, int tagI, TYPE *out, TYPE **arg) {{

    if (tagI==1) {{
        int tmp = ny;
        ny = nx;
        nx = tmp;
    }}

    return AssignZeroCpu_{self.gencode_filename}< TYPE > (nx, ny, out, arg);

}}

template < typename TYPE >
int launch_keops_cpu_{self.gencode_filename}(int dimY,
                                         int nx,
                                         int ny,
                                         int tagI,
                                         int tagZero,
                                         int use_half,
                                         int dimred,
                                         int use_chunk_mode,
                                         std::vector< int > indsi, std::vector< int > indsj, std::vector< int > indsp,
                                         int dimout,
                                         std::vector< int > dimsx, std::vector< int > dimsy, std::vector< int > dimsp,
                                         int **ranges,
                                         std::vector< int > shapeout, TYPE *out,
                                         TYPE **arg,
                                         std::vector< std::vector< int > > argshape) {{


    return launch_keops_{self.gencode_filename}< TYPE > (nx, ny, tagI, out, arg);

}}

                    """



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_for_loop,
    c_zero_float,
)
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////     Matrix-vector product      A x b                           ////
# /////////////////////////////////////////////////////////////////////////


class MatVecMult(Operation):
    string_id = "MatVecMult"

    def __init__(self, A, B):
        # A is vector of size n*p, interpreted as matrix, B is vector of size p, interpreted as column vector
        # output is vector of size n
        if A.dim % B.dim != 0:
            KeOps_Error(
                "Dimensions of A and B are not compatible for matrix-vector product"
            )
        super().__init__(A, B)
        self.dim = A.dim // B.dim

    def Op(self, out, table, inA, inB):
        q = c_variable("int")
        loop, i = c_for_loop(0, self.dim, 1, pragma_unroll=True)
        inner_loop, k = c_for_loop(0, inB.dim, 1, pragma_unroll=True)
        return f"""
                    #if C_CONTIGUOUS     // row major
                        {q.declare_assign(0)}
                        {loop(out[i].assign(c_zero_float) + inner_loop(out[i].add_assign(inA[q] * inB[k]) + q.add_assign(1)))}
                    #else               // column major
                        {loop(out[i].assign(c_zero_float) + inner_loop(out[i].add_assign(inA[k * self.dim + i] * inB[k])))}
                    #endif
                """

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths import TensorProd, VecMatMult

        A, B = self.children
        return A.DiffT(v, TensorProd(gradin, B)) + B.DiffT(v, VecMatMult(gradin, A))

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 2  # number of arguments
    test_argdims = [6, 2]  # dimensions of arguments for testing
    torch_op = None  # equivalent PyTorch operation



from keopscore.formulas.complex.Real2Complex import Real2Complex
from keopscore.formulas.complex.ComplexMult import ComplexMult
from keopscore.formulas.complex.ComplexSquareAbs import ComplexSquareAbs
from keopscore.formulas.complex.Conj import Conj
from keopscore.formulas.maths.Inv import Inv

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexDivide                           ////
# /////////////////////////////////////////////////////////////////////////


def ComplexDivide(f, g):
    return ComplexMult(Real2Complex(Inv(ComplexSquareAbs(g))), ComplexMult(f, Conj(g)))



from keopscore.formulas.Chunkable_Op import Chunkable_Op
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.code_gen_utils import c_zero_float, VectApply


##########################
######    Sum        #####
##########################


class Sum_Impl(Chunkable_Op):
    # the summation operation

    string_id = "Sum"

    dim = 1

    def Op(self, out, table, arg):
        return out.assign(c_zero_float) + VectApply(self.ScalarOp, out, arg)

    def ScalarOp(self, out, arg):
        return out.add_assign(arg)

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.SumT import SumT

        f = self.children[0]
        return f.DiffT(v, SumT(gradin, f.dim))

    def initacc_chunk(self, acc):
        return f"*{acc.id} = 0.0f;\n"

    def acc_chunk(self, acc, out):
        return f"*{acc.id} += *{out.id};\n"


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Sum(arg):
    if isinstance(arg, Zero):
        return Zero(1)
    else:
        return Sum_Impl(arg)



from keopscore.utils.code_gen_utils import (
    infinity,
    cast_to,
    c_zero_float,
    c_for_loop,
    c_variable,
    new_c_varname,
    c_if,
    c_array,
    use_pragma_unroll,
)
from keopscore.formulas.reductions.Reduction import Reduction
from keopscore.utils.misc_utils import KeOps_Error


class KMin_ArgKMin_Reduction(Reduction):
    """Implements the k-min-arg-k-min reduction operation : for each i or each j, find the
    values and indices of the k minimal values of Fij operation is vectorized: if Fij is vector-valued,
     arg-k-min is computed for each dimension."""

    string_id = "KMin_ArgKMin_Reduction"

    def __init__(self, formula, K, tagIJ):
        super().__init__(formula, tagIJ)

        self.K = K

        # dim is dimension of output of reduction ; for a arg-k-min reduction it is equal to the dimension of output of formula
        self.dim = 2 * K * formula.dim

        # We work with a (values,indices) vector
        self.dimred = self.dim  # dimension of inner reduction variables

    def InitializeReduction(self, acc):
        # Returns C++ code to be used at initialization phase of the reduction.
        if acc.dtype == "half2":
            KeOps_Error("not implemented")
        fdim, K = self.formula.dim, self.K
        outer_loop, k = c_for_loop(0, fdim, 1, pragma_unroll=True)
        inner_loop, l = c_for_loop(k, k + (2 * K * fdim), 2 * fdim, pragma_unroll=True)
        return outer_loop(
            inner_loop(
                acc[l].assign(infinity(acc.dtype)) + acc[l + fdim].assign(c_zero_float)
            )
        )

    def ReducePair(self, acc, xi):
        # Returns C++ code that implements the update phase of the reduction.
        dtype = xi.dtype
        fdim = self.formula.dim
        out = c_array(dtype, self.dimred, new_c_varname("out"))
        outer_loop, k = c_for_loop(0, fdim, 1)
        p = c_variable("int", new_c_varname("p"))
        q = c_variable("int", new_c_varname("q"))
        inner_loop, l = c_for_loop(k, self.dimred, 2 * fdim)
        inner_body = c_if(
            xi[p] < acc[q],
            out[l].assign(xi[p])
            + out[l + fdim].assign(xi[p + fdim])
            + p.add_assign(2 * fdim),
            out[l].assign(acc[q])
            + out[l + fdim].assign(acc[q + fdim])
            + q.add_assign(2 * fdim),
        )
        outer_body = p.declare_assign(k) + q.declare_assign(k) + inner_loop(inner_body)
        final_loop, k = c_for_loop(0, self.dimred, 1)
        return (
            out.declare() + outer_loop(outer_body) + final_loop(acc[k].assign(out[k]))
        )

    def ReducePairShort(self, acc, xi, ind):
        fdim, K = self.formula.dim, self.K
        dtype = xi.dtype

        xik = c_variable(dtype, new_c_varname("xik"))
        l = c_variable("int", new_c_varname("l"))
        k = c_variable("int", new_c_varname("k"))
        tmpl = c_variable(dtype, new_c_varname("tmpl"))
        indtmpl = c_variable("int", new_c_varname("indtmpl"))
        return f"""
                    {{
                        {xik.declare()}
                        {l.declare()}
                        {use_pragma_unroll()}
                        for(int {k.id}=0; {k.id}<{fdim}; {k.id}++) {{
                            {xik.assign(xi[k])}
                            {use_pragma_unroll()}                 
                            for({l.id}={(k+(K-1)*2*fdim).id}; {l.id}>={k.id} && {(xik<acc[l]).id}; {l.id}-={2*fdim}) {{
                                {tmpl.declare_assign(acc[l])}
                                {indtmpl.declare_assign(acc[l+fdim])}
                                {acc[l].assign(xik)}
                                {acc[l+fdim].assign(ind)}                      
                                if({l.id}<{(k+(2*fdim*(K-1))).id}) {{
                                    {acc[l+2*fdim].assign(tmpl)}
                                    {acc[l+2*fdim+fdim].assign(indtmpl)}
                                }}
                            }}
                        }}
                    }}
                """



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_log


class Log(VectorizedScalarOp):
    """the logarithm vectorized operation"""

    string_id = "Log"

    ScalarOpFun = keops_log

    @staticmethod
    def Derivative(f):
        return 1 / f

    # parameters for testing the operation (optional)
    test_ranges = [(0, 2)]  # range of argument



from keopscore.formulas.Operation import Operation
from keopscore.formulas.variables.Var import Var
from keopscore.config.chunks import enable_chunk, dim_treshold_chunk, specdims_use_chunk


class Chunkable_Op(Operation):
    def chunked_version(self, dimchk):
        chunked_args = [child.chunked_version(dimchk) for child in self.children]
        if self.params == ():
            return type(self)(*chunked_args)
        else:
            return type(self)(*chunked_args, params=self.params)

    def chunked_vars(self, cat):
        res = set()
        for child in self.children:
            res = set.union(res, set(child.chunked_vars(cat)))
        return list(res)

    def notchunked_vars(self, cat):
        return []

    @property
    def use_chunk(self):
        test = enable_chunk & all(child.is_chunkable for child in self.children)
        child = self.children[0]
        subtest = (child.dim >= dim_treshold_chunk) | (child.dim in specdims_use_chunk)
        test &= subtest
        return test

    def chunked_formula(self, dimchk):
        if self.use_chunk:
            return dict(
                formula=self.chunked_version(dimchk), dim_org=self.children[0].dim
            )
        else:
            return None

    def chunked_formulas(self, dimchk):
        chk_f = self.chunked_formula(dimchk)
        res = [chk_f] if chk_f is not None else []
        for child in self.children:
            res += child.chunked_formulas(dimchk)
        return res

    @property
    def num_chunked_formulas(self):
        return (
            sum(child.num_chunked_formulas for child in self.children) + self.use_chunk
        )

    def post_chunk_formula(self, ind):
        if self.use_chunk:
            return Var(ind, 1, 3)
        else:
            return type(self)(
                *(child.post_chunk_formula(ind) for child in self.children),
                params=self.params
            )



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.IntInv import IntInv
from keopscore.formulas.maths.Rsqrt import Rsqrt
from keopscore.utils.math_functions import keops_sqrt


##########################
######    Sqrt       #####
##########################


class Sqrt(VectorizedScalarOp):
    """the square root vectorized operation"""

    string_id = "Sqrt"

    ScalarOpFun = keops_sqrt

    @staticmethod
    def Derivative(f):
        return IntInv(2) * Rsqrt(f)

    # parameters for testing the operation (optional)
    test_ranges = [(0, 2)]  # range of argument



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_zero_float, c_for_loop
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexSum                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexSum(Operation):

    string_id = "ComplexSum"

    def __init__(self, f):
        if f.dim % 2 != 0:
            KeOps_Error("Dimension of F must be even")
        self.dim = 2
        super().__init__(f)

    def Op(self, out, table, inF):
        f = self.children[0]
        string = out[0].assign(c_zero_float)
        string += out[1].assign(out[0])
        forloop, i = c_for_loop(0, f.dim, 2, pragma_unroll=True)
        body = out[0].add_assign(inF[i])
        body += out[1].add_assign(inF[i + 1])
        string += forloop(body)
        return string

    def DiffT(self, v, gradin):
        from keopscore.formulas.complex.ComplexSumT import ComplexSumT

        f = self.children[0]
        return f.DiffT(v, ComplexSumT(gradin, f.dim))



from keopscore import cuda_block_size
from keopscore.config.chunks import dimchunk
from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.formulas.reductions.sum_schemes import *
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    load_vars,
    load_vars_chunks,
    load_vars_chunks_offsets,
    sizeof,
    pointer,
    table,
    table4,
    Var_loader,
    use_pragma_unroll,
)
from keopscore.mapreduce.Chunk_Mode_Constants import Chunk_Mode_Constants


def do_chunk_sub_ranges(
    dtype,
    red_formula,
    fun_chunked_curr,
    dimchunk_curr,
    dimsx,
    dimsy,
    indsi,
    indsj,
    indsi_chunked,
    indsj_chunked,
    acc,
    tile,
    i,
    j,
    jstart,
    start_y,
    chunk,
    end_x,
    end_y,
    nbatchdims,
    indices_i,
    indices_j,
    arg,
    fout,
    xi,
    yj,
    yjrel,
    param_loc,
):
    chk = Chunk_Mode_Constants(red_formula)
    fout_tmp_chunk = c_array(dtype, chk.fun_chunked.dim)
    xiloc = c_variable(pointer(dtype), f"({xi.id} + {chk.dimx_notchunked})")
    yjloc = c_variable(
        pointer(dtype), f"({yj.id} + threadIdx.x * {chk.dimy} + {chk.dimy_notchunked})"
    )

    load_chunks_routine_i = load_vars_chunks(
        indsi_chunked,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        xiloc,
        arg,
        chunk,
        row_index=i,
    )

    varloader_global = Var_loader(red_formula)
    indsi_global = varloader_global.indsi
    indsj_global = varloader_global.indsj

    threadIdx_x = c_variable("int", "threadIdx.x")
    load_chunks_routine_i_batches = load_vars_chunks_offsets(
        indsi_chunked,
        indsi_global,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        xiloc,
        arg,
        chunk,
        indices_i,
        row_index=threadIdx_x,
    )

    load_chunks_routine_j = load_vars_chunks(
        indsj_chunked,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        yjloc,
        arg,
        chunk,
        row_index=j,
    )

    load_chunks_routine_j_batches = load_vars_chunks_offsets(
        indsj_chunked,
        indsj_global,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        yjloc,
        arg,
        chunk,
        indices_j,
        row_index=j - start_y,
    )

    chktable = table(
        chk.nminargs,
        dimsx,
        dimsy,
        chk.dimsp,
        indsi,
        indsj,
        chk.indsp,
        xi,
        yjrel,
        param_loc,
    )
    foutj = c_variable(pointer(dtype), "foutj")

    return f"""
                {fout_tmp_chunk.declare()}
                if ({i.id} < {end_x.id}) {{
                    if ({nbatchdims.id}==0) {{
                        {load_chunks_routine_i}
                    }} else {{
                        {load_chunks_routine_i_batches}
                    }}
                }} 
                if ({j.id} < {end_y.id}) {{ // we load yj from device global memory only if j<ny
                    if ({nbatchdims.id}==0) {{
                        {load_chunks_routine_j}
                    }} else {{
                        {load_chunks_routine_j_batches}
                    }}
                }}
                __syncthreads();
                if ({i.id} < {end_x.id}) {{ // we compute only if needed
            		{dtype} *yjrel = {yj.id}; // Loop on the columns of the current block.
            		for (int jrel = 0; (jrel < blockDim.x) && (jrel < {end_y.id} - jstart); jrel++, yjrel += {chk.dimy}) {{
            			{dtype} *foutj = {fout.id} + jrel*{chk.fun_chunked.dim};
                        {fun_chunked_curr(fout_tmp_chunk, chktable)}
                        {chk.fun_chunked.acc_chunk(foutj, fout_tmp_chunk)}
            		}}
                }} 
                __syncthreads();
            """


class GpuReduc1D_ranges_chunks(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.chk = Chunk_Mode_Constants(self.red_formula)
        self.dimy = self.chk.dimy
        self.blocksize_chunks = min(
            cuda_block_size, 1024, 49152 // max(1, self.dimy * sizeof(self.dtype))
        )

    def get_code(self):

        super().get_code()

        red_formula = self.red_formula
        dtype = self.dtype
        dtypeacc = self.dtypeacc

        varloader_global = Var_loader(red_formula)
        indsi_global = varloader_global.indsi
        indsj_global = varloader_global.indsj

        i = self.i
        j = self.j

        arg = self.arg
        args = self.args

        nvarsi, nvarsj, nvarsp = (
            len(self.varloader.Varsi),
            len(self.varloader.Varsj),
            len(self.varloader.Varsp),
        )
        nvars = nvarsi + nvarsj + nvarsp

        indices_i = c_array("int", nvarsi, "indices_i")
        indices_j = c_array("int", nvarsj, "indices_j")
        indices_p = c_array("int", nvarsp, "indices_p")

        declare_assign_indices_i = "int *indices_i = offsets;" if nvarsi > 0 else ""
        declare_assign_indices_j = (
            f"int *indices_j = offsets + {nvarsi};" if nvarsj > 0 else ""
        )
        declare_assign_indices_p = (
            f"int *indices_p = offsets + {nvarsi} + {nvarsj};" if nvarsp > 0 else ""
        )

        yjrel = c_array(dtype, varloader_global.dimy, "yjrel")

        jreltile = c_variable("int", "(jrel + tile * blockDim.x)")

        chk = self.chk
        param_loc = c_array(dtype, chk.dimp, "param_loc")
        acc = c_array(dtypeacc, chk.dimred, "acc")
        sum_scheme = eval(self.sum_scheme_string)(red_formula, dtype, dimred=chk.dimred)
        xi = c_array(dtype, chk.dimx, "xi")
        fout_chunk = c_array(
            dtype, self.blocksize_chunks * chk.dimout_chunk, "fout_chunk"
        )
        yj = c_variable(pointer(dtype), "yj")
        yjloc = c_array(dtype, chk.dimy, f"(yj + threadIdx.x * {chk.dimy})")

        fout_chunk_loc = c_variable(
            pointer(dtype), f"({fout_chunk.id}+jrel*{chk.dimout_chunk})"
        )

        tile = c_variable("int", "tile")
        nx = c_variable("int", "nx")
        ny = c_variable("int", "ny")

        jstart = c_variable("int", "jstart")
        chunk = c_variable("int", "chunk")

        end_x = c_variable("int", "end_x")
        end_y = c_variable("int", "end_y")

        starty = c_variable("int", "start_y")

        nbatchdims = c_variable("int", "nbatchdims")

        chunk_sub_routine = do_chunk_sub_ranges(
            dtype,
            red_formula,
            chk.fun_chunked,
            dimchunk,
            chk.dimsx,
            chk.dimsy,
            chk.indsi,
            chk.indsj,
            chk.indsi_chunked,
            chk.indsj_chunked,
            acc,
            tile,
            i,
            j,
            jstart,
            starty,
            chunk,
            end_x,
            end_y,
            nbatchdims,
            indices_i,
            indices_j,
            arg,
            fout_chunk,
            xi,
            yj,
            yjrel,
            param_loc,
        )

        last_chunk = c_variable("int", f"{chk.nchunks-1}")
        chunk_sub_routine_last = do_chunk_sub_ranges(
            dtype,
            red_formula,
            chk.fun_lastchunked,
            chk.dimlastchunk,
            chk.dimsx_last,
            chk.dimsy_last,
            chk.indsi,
            chk.indsj,
            chk.indsi_lastchunked,
            chk.indsj_lastchunked,
            acc,
            tile,
            i,
            j,
            jstart,
            starty,
            last_chunk,
            end_x,
            end_y,
            nbatchdims,
            indices_i,
            indices_j,
            arg,
            fout_chunk,
            xi,
            yj,
            yjrel,
            param_loc,
        )

        foutj = c_array(dtype, chk.dimout_chunk, "foutj")

        chktable_out = table4(
            chk.nminargs + 1,
            chk.dimsx,
            chk.dimsy,
            chk.dimsp,
            [chk.dimout_chunk],
            chk.indsi,
            chk.indsj,
            chk.indsp,
            [chk.nminargs],
            xi,
            yjrel,
            param_loc,
            foutj,
        )
        fout_tmp = c_array(dtype, chk.dimfout, "fout_tmp")
        outi = c_array(dtype, chk.dimout, f"(out + i * {chk.dimout})")

        threadIdx_x = c_variable("int", "threadIdx.x")

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void GpuConv1DOnDevice_ranges(int nx, int ny, int nbatchdims,
                                                    int *offsets_d, int *lookup_d, int *slices_x,
                                                    int *ranges_y, {dtype} *out, {dtype} **{arg.id}) {{
                                                        
                          int offsets[{nvars}];
                          {declare_assign_indices_i}
                          {declare_assign_indices_j}
                          {declare_assign_indices_p}
                          
                          if (nbatchdims > 0)
                              for (int k = 0; k < {nvars}; k++)
                                  offsets[k] = offsets_d[ {nvars} * blockIdx.x + k ];
                                  
                          // Retrieve our position along the laaaaarge [1,~nx] axis: -----------------
                          int range_id= (lookup_d)[3*blockIdx.x] ;
                          int start_x = (lookup_d)[3*blockIdx.x+1] ;
                          int end_x   = (lookup_d)[3*blockIdx.x+2] ;
                          
                          // The "slices_x" vector encodes a set of cutting points in
                          // the "ranges_y" array of ranges.
                          // As discussed in the Genred docstring, the first "0" is implicit:
                          int start_slice = range_id < 1 ? 0 : slices_x[range_id-1];
                          int end_slice   = slices_x[range_id];
                          
                          // get the index of the current thread
                          int i = start_x + threadIdx.x;
                          
                          // declare shared mem
                          extern __shared__ {dtype} yj[];

                          // load parameters variables from global memory to local thread memory
                          {param_loc.declare()}
                          if (nbatchdims == 0) {{
                              {load_vars(chk.dimsp, chk.indsp, param_loc, args)}
                          }} else {{
                              {load_vars(chk.dimsp, chk.indsp, param_loc, args, offsets=indices_p)}
                          }}
                          
                          {acc.declare()}
                          
                          {sum_scheme.declare_temporary_accumulator()}                     
                          
                          if (i < end_x) {{
                            {red_formula.InitializeReduction(acc)} // acc = 0
                            {sum_scheme.initialize_temporary_accumulator_first_init()}
                          }}

                          {xi.declare()}

                          {fout_chunk.declare()}
                          
                          if (i < end_x) {{
                              // load xi variables from global memory to local thread memory
                              if (nbatchdims == 0) {{
                                  {load_vars(chk.dimsx_notchunked, chk.indsi_notchunked, xi, args, row_index=i)} 
                              }} else {{
                                  {load_vars(chk.dimsx_notchunked, chk.indsi_notchunked, xi, args, 
                                              row_index=threadIdx_x, offsets=indices_i, indsref=indsi_global)}
                              }} 
                          }}
                          
                          int start_y = ranges_y[2*start_slice], end_y = 0;
                          for( int index = start_slice ; index < end_slice ; index++ ) {{
                              if( (index+1 >= end_slice) || (ranges_y[2*index+2] != ranges_y[2*index+1]) ) {{
                                  //start_y = ranges_y[2*index] ;
                                  end_y = ranges_y[2*index+1];
                                  
                                  for(int jstart = start_y, tile = 0; jstart < end_y; jstart += blockDim.x, tile++) {{
                                      // get the current column
                                      int j = jstart + threadIdx.x;
                                      
                                      if(j<end_y) // we load yj from device global memory only if j<end_y
                                          if (nbatchdims == 0) {{
                                              // load yj variables from global memory to shared memory
                                              {load_vars(chk.dimsy_notchunked, chk.indsj_notchunked, yjloc, args, row_index=j)} 
                                          }} else {{
                                              // Possibly, with offsets as we support broadcasting over batch dimensions
                                              {load_vars(chk.dimsy_notchunked, chk.indsj_notchunked, yjloc, args, 
                                                          row_index=j-starty, offsets=indices_j, indsref=indsj_global)}
                                          }}
                                      __syncthreads();
                                      
                                      if(i<end_x) {{ // we compute x1i only if needed
                                      for (int jrel = 0; (jrel < blockDim.x) && (jrel < end_y - jstart); jrel++) {{
                                          {chk.fun_chunked.initacc_chunk(fout_chunk_loc)}
                                      }}
                                      {sum_scheme.initialize_temporary_accumulator_block_init()}
                                  }}
                                  
                                  // looping on chunks (except the last)
                          		  {use_pragma_unroll()}
                          		  for (int chunk=0; chunk<{chk.nchunks}-1; chunk++) {{
                                      {chunk_sub_routine}
                                  }}
                                  // last chunk
                                  {chunk_sub_routine_last}
                                  
                                  if (i < end_x) {{
                                      {dtype} * yjrel = yj; // Loop on the columns of the current block.
                                      if (nbatchdims == 0) {{
                                          for (int jrel = 0; (jrel < blockDim.x) && (jrel <end_y - jstart); jrel++, yjrel += {chk.dimy}) {{
                                              {dtype} *foutj = fout_chunk + jrel*{chk.dimout_chunk};
                                              {fout_tmp.declare()}
                                              {chk.fun_postchunk(fout_tmp, chktable_out)}
                                              {sum_scheme.accumulate_result(acc, fout_tmp, jreltile)}
                                          }}
                                      }} else {{
                                          for (int jrel = 0; (jrel < blockDim.x) && (jrel <end_y - jstart); jrel++, yjrel += {chk.dimy}) {{
                                              {dtype} *foutj = fout_chunk + jrel*{chk.dimout_chunk};
                                              {fout_tmp.declare()}
                                              {chk.fun_postchunk(fout_tmp, chktable_out)}
                                              {sum_scheme.accumulate_result(acc, fout_tmp, jreltile)}
                                          }}
                                      }}
                                      {sum_scheme.final_operation(acc)}
                                  }}
                                  __syncthreads();
                              }}
                              
                              if(index+1 < end_slice) {{
                                  start_y = ranges_y[2*index+2] ;
                  			  }}
                          }}
                      }}
                      if (i < end_x) {{
                          {red_formula.FinalizeOutput(acc, outi, i)} 
                      }}
                  }}
                    """



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_atan


class Atan(VectorizedScalarOp):
    """the arc-tangent vectorized operation"""

    string_id = "Atan"

    ScalarOpFun = keops_atan

    @staticmethod
    def Derivative(f):
        return 1 / (1 + f**2)



from keopscore.formulas.Operation import Broadcast
from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Mult import Mult_Impl
from keopscore.formulas.variables.IntCst import IntCst, IntCst_Impl
from keopscore.formulas.variables.Zero import Zero


##########################
######    Add        #####
##########################


class Add_Impl(VectorizedScalarOp):
    """the binary addition operation"""

    string_id = "Add"
    print_spec = "+", "mid", 4

    def ScalarOp(self, out, arg0, arg1):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        return f"{out.id} = {arg0.id}+{arg1.id};\n"

    def DiffT(self, v, gradin):
        fa, fb = self.children
        return fa.DiffT(v, gradin) + fb.DiffT(v, gradin)

    # parameters for testing the operation (optional)
    nargs = 2  # number of arguments


# N.B. The following separate function could theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Add(arg0, arg1):
    if isinstance(arg0, Zero):
        return Broadcast(arg1, arg0.dim)
    elif isinstance(arg1, Zero):
        return Broadcast(arg0, arg1.dim)
    elif arg0 == arg1:
        return IntCst(2) * arg0
    elif isinstance(arg0, Mult_Impl) and isinstance(arg0.children[0], IntCst_Impl):
        if arg0.children[1] == arg1:
            #  factorization :  n*x + x = (n+1)*x
            return IntCst(arg0.children[0].val + 1) * arg1
        elif (
            isinstance(arg1, Mult_Impl)
            and isinstance(arg1.children[0], IntCst_Impl)
            and arg1.children[1] == arg0.children[1]
        ):
            #  factorization :  m*x + n*x = (m+n)*x
            return (
                IntCst(arg0.children[0].val + arg1.children[0].val) * arg0.children[1]
            )
    if (
        isinstance(arg1, Mult_Impl)
        and isinstance(arg1.children[0], IntCst_Impl)
        and arg1.children[1] == arg0
    ):
        #  factorization :  x + n*x = (n+1)*x
        return IntCst(arg1.children[0].val + 1) * arg0
    return Add_Impl(arg0, arg1)



from keopscore import cuda_block_size
from keopscore.config.chunks import dimfinalchunk
from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.formulas.reductions.Sum_Reduction import Sum_Reduction
from keopscore.formulas.reductions.sum_schemes import *
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    load_vars,
    load_vars_chunks,
    sizeof,
    pointer,
    Var_loader,
    use_pragma_unroll,
)
from keopscore.utils.misc_utils import KeOps_Error


def do_finalchunk_sub(
    dtype,
    varfinal,
    dimfinalchunk_curr,
    acc,
    i,
    j,
    jstart,
    chunk,
    nx,
    ny,
    arg,
    fout,
    yj,
    out,
):

    dimout = varfinal.dim
    yjloc = c_variable(pointer(dtype), f"({yj.id} + threadIdx.x * {dimfinalchunk})")
    load_chunks_routine_j = load_vars_chunks(
        [varfinal.ind],
        dimfinalchunk,
        dimfinalchunk_curr,
        varfinal.dim,
        yjloc,
        arg,
        chunk,
        row_index=j,
    )
    return f"""
                {acc.assign(c_zero_float)}
                {dtype} *yjrel = yj;
                if ({j.id} < {ny.id}) {{ // we load yj from device global memory only if j<ny
                    {load_chunks_routine_j}
                }}
                __syncthreads();
                for (int jrel = 0; (jrel < blockDim.x) && (jrel < {ny.id} - {jstart.id}); jrel++, yjrel += {dimfinalchunk}) {{          
                    if ({i.id} < {nx.id}) {{ // we compute only if needed
                        {use_pragma_unroll()}
                        for (int k=0; k<{dimfinalchunk_curr}; k++) {{
                            {acc.id}[k] += yjrel[k] * fout[jrel];
                        }}
                    }}
                    __syncthreads();
                }}
                if ({i.id} < {nx.id}) {{
                    {use_pragma_unroll()}
                    for (int k=0; k<{dimfinalchunk_curr}; k++)
                        {out.id}[i*{dimout}+{chunk.id}*{dimfinalchunk}+k] += {acc.id}[k];
                }}
                __syncthreads();
            """


class GpuReduc1D_finalchunks(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)

    def get_code(self):

        super().get_code()
        dtype = self.dtype
        dtypeacc = self.dtypeacc
        i = self.i
        j = self.j
        nx = c_variable("int", "nx")
        ny = c_variable("int", "ny")
        jstart = c_variable("int", "jstart")
        chunk = c_variable("int", "chunk")
        arg = self.arg
        args = self.args
        yj = c_variable(pointer(dtype), "yj")
        out = c_variable(pointer(dtype), "out")

        fun_internal = Sum_Reduction(
            self.red_formula.formula.children[0], self.red_formula.tagI
        )
        formula = fun_internal.formula

        varfinal = self.red_formula.formula.children[1]
        nchunks = 1 + (varfinal.dim - 1) // dimfinalchunk
        dimlastfinalchunk = varfinal.dim - (nchunks - 1) * dimfinalchunk
        varloader = Var_loader(fun_internal)
        dimsx = varloader.dimsx
        dimsy = varloader.dimsy
        dimsp = varloader.dimsp
        indsi = varloader.indsi
        indsj = varloader.indsj
        indsp = varloader.indsp
        dimx = sum(dimsx)
        dimy = sum(dimsy)
        dimp = sum(dimsp)
        dimout = varfinal.dim
        dimfout = fun_internal.formula.dim
        if dimfout != 1:
            KeOps_Error("dimfout should be 1")
        sum_scheme = self.sum_scheme

        self.dimy = max(dimfinalchunk, dimy)
        blocksize_chunks = min(
            cuda_block_size, 1024, 49152 // max(1, self.dimy * sizeof(self.dtype))
        )

        if not isinstance(sum_scheme, block_sum):
            KeOps_Error("only block_sum available")
        param_loc = c_array(dtype, dimp, "param_loc")
        fout = c_array(dtype, dimfout * blocksize_chunks, "fout")
        xi = c_array(dtype, dimx, "xi")
        acc = c_array(dtypeacc, dimfinalchunk, "acc")
        yjloc = c_array(dtype, dimy, f"(yj + threadIdx.x * {dimy})")
        foutjrel = c_array(dtype, dimfout, f"({fout.id}+jrel*{dimfout})")
        yjrel = c_array(dtype, dimy, "yjrel")
        table = self.varloader.table(xi, yjrel, param_loc)

        last_chunk = c_variable("int", f"{nchunks-1}")

        chunk_sub_routine = do_finalchunk_sub(
            dtype,
            varfinal,
            dimfinalchunk,
            acc,
            i,
            j,
            jstart,
            chunk,
            nx,
            ny,
            arg,
            fout,
            yj,
            out,
        )

        chunk_sub_routine_last = do_finalchunk_sub(
            dtype,
            varfinal,
            dimlastfinalchunk,
            acc,
            i,
            j,
            jstart,
            last_chunk,
            nx,
            ny,
            arg,
            fout,
            yj,
            out,
        )

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void GpuConv1DOnDevice(int nx, int ny, {dtype} *out, {dtype} **{arg.id}) {{
    
                          // get the index of the current thread
                          int i = blockIdx.x * blockDim.x + threadIdx.x;

                          // declare shared mem
                          extern __shared__ {dtype} yj[];
            
                          // load parameter(s)
                          {param_loc.declare()}
                          {load_vars(dimsp, indsp, param_loc, args)}
                          
                          {fout.declare()}
    
                          // get the value of variable (index with i)
                          {xi.declare()}
                          if (i < nx) {{
                              {load_vars(dimsx, indsi, xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                              {use_pragma_unroll()}
                              for (int k=0; k<{dimout}; k++) {{
                                  out[i*{dimout}+k] = 0.0f;
                              }}
                          }}
                          
                          {acc.declare()}

                          for (int jstart = 0, tile = 0; jstart < ny; jstart += blockDim.x, tile++) {{

                              // get the current column
                              int j = tile * blockDim.x + threadIdx.x;
                              if (j < ny) {{ // we load yj from device global memory only if j<ny
                                  {load_vars(dimsy, indsj, yjloc, args, row_index=j)} // load yj variables from global memory to shared memory
                              }}
                              __syncthreads();

                              if (i < nx) {{ // we compute x1i only if needed
                                  {dtype} * yjrel = yj; // Loop on the columns of the current block.
                                  for (int jrel = 0; (jrel < {blocksize_chunks}) && (jrel < ny - jstart); jrel++, yjrel += {dimy}) {{
                                      {formula(foutjrel, table)} // Call the function, which outputs results in fout
                                  }}
                              }}
        
                              __syncthreads();
        
                              for (int chunk=0; chunk<{nchunks-1}; chunk++) {{
                                  {chunk_sub_routine}
                              }}
                              {chunk_sub_routine_last}
                          }}
                        }}
                    """



from keopscore.utils.code_gen_utils import VectCopy
from keopscore.formulas.Operation import Operation


#######################
## Var operation
#######################


class Var(Operation):
    """Var operation class. Var(ind,dim,cat) is a symbolic
    object that encodes an input tensor in the call to the
    KeOps routine, where
    - ind gives the position of the input tensor in the list of tensors sent to the routine
    - dim gives the "dimension" of the data : each input tensor is interpreted as a matrix
    of size (n,dim), where n is dynamically handled and dim is known at compile time.
    - cat is the "category" of the variable : either a "i"-indexed variable (cat=0),
    a "j"-indexed variable (cat=1), or a parameter variable (cat=2)"""

    string_id = "Var"

    def __init__(self, ind, dim, cat):
        super().__init__()
        self.ind = ind
        self.dim = dim
        self.cat = cat
        self.Vars_ = {self}
        self.params = (ind, dim, cat)

    # custom __eq__ and __hash__ methods, required to handle properly the union of two sets of Var objects
    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.ind == other.ind
            and self.dim == other.dim
            and self.cat == other.cat
        )

    def __hash__(self):
        return hash((self.ind, self.dim, self.cat))

    def Op(self, out, table):
        return VectCopy(out, table[self.ind])

    # Assuming that the gradient wrt. Var is GRADIN, how does it affect V ?
    # Var::DiffT<V, grad_input> = grad_input   if V == Var (in the sense that it represents the same symb. var.)
    #                             Zero(V::DIM) otherwise
    def DiffT(self, v, gradin):
        from keopscore.formulas.variables.Zero import Zero

        return gradin if v == self else Zero(v.dim)

    def chunked_version(self, dimchk):
        return Var(self.ind, dimchk, self.cat)

    @property
    def is_chunkable(self):
        return True

    def chunked_formulas(self, dimchk):
        return []

    @property
    def num_chunked_formulas(self):
        return 0

    def post_chunk_formula(self, ind):
        return Var(self.ind, self.dim, self.cat)

    def chunked_vars(self, cat):
        return self.Vars(cat)

    def notchunked_vars(self, cat):
        return set()



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import value
from keopscore.utils.misc_utils import KeOps_Error

############################
######    ELEMENT EXTRACTION : Elem(f,m) (aka get_item)       #####
############################


class Elem(Operation):
    string_id = "Elem"

    def __init__(self, f, m):
        super().__init__(f, params=(m,))
        if f.dim <= m:
            KeOps_Error("Index out of bound in Elem")
        self.dim = 1
        self.m = m

    def Op(self, out, table, arg):
        return value(out).assign(arg[self.m])

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.ElemT import ElemT

        f = self.children[0]
        return f.DiffT(v, ElemT(gradin, f.dim, self.m))

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [5]  # dimensions of arguments for testing
    test_params = [3]  # dimensions of parameters for testing
    torch_op = "lambda x, m : x[..., m][..., None]"



from keopscore.utils.code_gen_utils import VectApply
from keopscore.formulas.Operation import Operation
from keopscore.utils.misc_utils import KeOps_Error


class VectorizedScalarOp(Operation):
    # class for operations that are vectorized or broadcasted
    # scalar operations,
    # such as Exp(f), Cos(f), Mult(f,g), Subtract(f,g), etc.

    def __init__(self, *args, params=()):
        dims = set(arg.dim for arg in args)
        if len(dims) > 2 or (len(dims) == 2 and min(dims) != 1):
            KeOps_Error("dimensions are not compatible for VectorizedScalarOp")
        super().__init__(*args, params=params)

    @property
    def dim(self):
        # dim gives the output dimension of the operation,
        # here it is the same as the output dimension of the child operation
        return max(child.dim for child in self.children)

    def Op(self, out, table, *args):
        # Atomic evaluation of the operation : it consists in a simple
        # for loop around the call to the correponding scalar operation
        return VectApply(self.ScalarOp, out, *args)

    def ScalarOp(self, out, *args):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        return out.assign(type(self).ScalarOpFun(*args, *self.params))

    def DiffT(self, v, gradin):
        derivatives = self.Derivative(*self.children, *self.params)
        if len(self.children) == 1:
            derivatives = (derivatives,)
        return sum(f.DiffT(v, gradin * df) for f, df in zip(self.children, derivatives))

    def Derivative(self):
        pass

    @property
    def is_chunkable(self):
        return all(child.is_chunkable for child in self.children)

    def chunked_version(self, dimchk):
        return type(self)(
            *(
                (child if child.dim == 1 else child.chunked_version(dimchk))
                for child in self.children
            ),
            *self.params
        )

    def chunked_vars(self, cat):
        res = set()
        for child in self.children:
            if child.dim != 1:
                res = set.union(res, set(child.chunked_vars(cat)))
        return list(res)

    def notchunked_vars(self, cat):
        res = set()
        for child in self.children:
            if child.dim == 1:
                res = set.union(res, set(child.Vars(cat)))
            else:
                res = set.union(res, set(child.notchunked_vars(cat)))
        return list(res)

    enable_test = True



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.Log import Log
from keopscore.utils.math_functions import keops_xlogx


class XLogX(VectorizedScalarOp):
    """the x*log(x) vectorized operation"""

    string_id = "XLogX"

    ScalarOpFun = keops_xlogx

    @staticmethod
    def Derivative(f):
        return Log(f) + 1

    # parameters for testing the operation (optional)
    test_ranges = [(0, 2)]  # range of argument
    torch_op = "lambda x:x*torch.log(x)"



import sys, os
from os import path

###########################################################
# Verbosity level
verbose = True
if os.getenv("KEOPS_VERBOSE") == "0":
    verbose = False

here = path.abspath(path.dirname(__file__))
with open(os.path.join(here, "keops_version"), encoding="utf-8") as v:
    __version__ = v.read().rstrip()

from .config.config import set_build_folder, get_build_folder
from .utils.code_gen_utils import clean_keops

# flags for debugging :
# prints information about atomic operations during code building
debug_ops = False
# adds C++ code for printing all input and output values
# for all atomic operations during computations
debug_ops_at_exec = False

cuda_block_size = 192

from . import config as keopscoreconfig

if keopscoreconfig.config.use_cuda:
    keopscoreconfig.config.init_cudalibs()
    from .binders.nvrtc.Gpu_link_compile import Gpu_link_compile
    from .binders.nvrtc.Gpu_link_compile import jit_compile_dll

    if not os.path.exists(jit_compile_dll()):
        Gpu_link_compile.compile_jit_compile_dll()



from keopscore.formulas.maths import Concat, Exp, Extract
from keopscore.formulas.reductions.Reduction import Reduction
from keopscore.formulas.reductions.Sum_Reduction import Sum_Reduction
from keopscore.formulas.variables.IntCst import IntCst
from keopscore.utils.code_gen_utils import (
    neg_infinity,
    c_zero_float,
    new_c_varname,
    c_variable,
    c_for_loop,
)
from keopscore.utils.math_functions import keops_exp
from keopscore.utils.misc_utils import KeOps_Error


class Max_SumShiftExpWeight_Reduction(Reduction):
    """Implements the coupled reduction operation m_i=max_j f_ij, s_i=sum_j exp(f_ij-m_i) g_ij
    where f and g are two formulas. f must be scalar-valued.
    This reduction is the base for numerically stable computation of log-sum-exp and softmax type reductions."""

    string_id = "Max_SumShiftExpWeight_Reduction"

    def __init__(self, formulaF, tagIJ, formulaG=IntCst(1)):
        if formulaF.dim != 1:
            KeOps_Error(
                "Max_SumShiftExpWeight_Reduction requires first formula of dimension 1."
            )
        super().__init__(Concat(formulaF, formulaG), tagIJ)
        self.formulaF = formulaF
        self.formulaG = formulaG
        self.dim = formulaF.dim + formulaG.dim  # dimension of final output of reduction
        self.dimred = self.dim  # dimension of inner reduction variables

    def InitializeReduction(self, acc):
        """Returns C++ code to be used at initialization phase of the reduction.
        We fill empty cells with the neutral element of the reduction operation,
         (-inf,0) = e^{-inf} * 0 = 0"""
        m, s = acc.split(1, self.formulaG.dim)
        return m.assign(neg_infinity(acc.dtype)) + s.assign(c_zero_float)

    def ReducePair(self, acc, xi):
        """Returns C++ code that implements the update phase of the reduction.
        (m,s) + (m',s'), i.e. exp(m)*s + exp(m')*s'"""

        if xi.dtype == "half2":
            KeOps_Error("Not implemented.")

        tmpexp = c_variable(acc.dtype, new_c_varname("tmpexp"))
        loop, k = c_for_loop(1, self.dimred, 1, pragma_unroll=True)
        return f"""       
                      {tmpexp.declare()}
                      if ({acc.id}[0] > {xi.id}[0]) {{ // =  exp(m)  * (s + s'*exp(m'-m))   if m > m'
                        {tmpexp.assign(keops_exp(xi[0]-acc[0]))}
                        {loop(acc[k].add_assign(xi[k]*tmpexp))}
                      }} else {{             // =  exp(m') * (s' + exp(m-m')*s)   if m <= m'
                        {tmpexp.assign(keops_exp(acc[0]-xi[0]))}
                        {loop(acc[k].assign(xi[k]+tmpexp*acc[k]))}
                        {acc[0].assign(xi[0])}
                      }}
              """

    def ReducePairShort(self, acc, xi, ind):
        return self.ReducePair(acc, xi)

    def KahanScheme(self, acc, xi, tmp):
        if xi.dtype == "half2":
            KeOps_Error("Not implemented.")
        tmpexp = c_variable(acc.dtype, new_c_varname("tmpexp"))
        loop, k = c_for_loop(1, self.dimred, 1, pragma_unroll=True)
        a = c_variable(acc.dtype, new_c_varname("a"))
        b = c_variable(acc.dtype, new_c_varname("b"))
        u = c_variable(acc.dtype, new_c_varname("u"))
        return f"""
                        {tmpexp.declare()}
                        if ({acc.id}[0] > {xi.id}[0])    // =  exp(m)  * (s + s'*exp(m'-m))   if m > m'
                        {{      
                            {tmpexp.assign(keops_exp(xi[0]-acc[0]))}
                            {loop( a.declare_assign(xi[k]*tmpexp-tmp[k-1]) + b.declare_assign(acc[k]+a) + tmp[k-1].assign((b-acc[k])-a) + acc[k].assign(b))}
                        }} 
                        else      // =  exp(m') * (s' + exp(m-m')*s)   if m <= m'
                        {{             
                            {tmpexp.assign(keops_exp(acc[0]-xi[0]))}
                            {loop( u.declare_assign(tmpexp*acc[k]) + a.declare_assign(xi[k]-tmpexp*tmp[k-1]) + b.declare_assign(u+a) + tmp[k-1].assign((b-u)-a) + acc[k].assign(b))}
                            {acc[0].assign(xi[0])}
                        }}
                """

    def DiffT(self, v, gradin, MS):
        """
        // Beware: the formula that we use for the gradient is *only* valid
        // if the output [M,S] = Max_SumShiftExp(F,G) has been flattened through a
        // L = M + log(S) (Log-Sum-Exp) or a weighted Soft-Max
        // operation (as done by the Python bindings), and if
        // GRADIN = [Grad(L), Grad(L)/S ]
        // has been backpropagated from L.
        """
        from keopscore.formulas.autodiff import Grad

        M = Extract(MS, 0, self.formulaF.dim)
        S = Extract(gradin, self.formulaF.dim, self.formulaG.dim)
        return Grad(
            Sum_Reduction(Exp(self.formulaF - M) * self.formulaG, self.tagI), v, S
        )


Max_SumShiftExp_Reduction = Max_SumShiftExpWeight_Reduction



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_for_loop,
    c_zero_float,
)
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////     Vector-matrix product           b x A                       ////
# /////////////////////////////////////////////////////////////////////////


class VecMatMult(Operation):
    string_id = "VecMatMult"

    def __init__(self, B, A):
        # A is vector of size n*p, interpreted as matrix, B is vector of size n, interpreted as row vector
        # output is vector of size p
        if A.dim % B.dim != 0:
            KeOps_Error(
                "Dimensions of A and B are not compatible for vector-matrix product"
            )
        super().__init__(B, A)
        self.dim = A.dim // B.dim

    def Op(self, out, table, inB, inA):
        q = c_variable("int")
        loop, i = c_for_loop(0, self.dim, 1, pragma_unroll=True)
        inner_loop, k = c_for_loop(0, inB.dim, 1, pragma_unroll=True)
        return f"""
                    #if C_CONTIGUOUS     // row major
                        {q.declare_assign(0)}
                        {loop(out[i].assign(c_zero_float) + inner_loop(out[i].add_assign(inA[k * self.dim + i] * inB[k])))}
                    #else               // column major
                        {q.declare_assign(0)}
                        {loop(out[i].assign(c_zero_float) + inner_loop(out[i].add_assign(inA[q] * inB[k]) + q.add_assign(1)))}
                    #endif
                """

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.MatVecMult import MatVecMult
        from keopscore.formulas.maths.TensorProd import TensorProd

        B, A = self.children
        return A.DiffT(v, TensorProd(B, gradin)) + B.DiffT(v, MatVecMult(A, gradin))



from keopscore import cuda_block_size
from keopscore.config.chunks import dimchunk
from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.formulas.reductions.sum_schemes import *
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    load_vars,
    load_vars_chunks,
    sizeof,
    pointer,
    table,
    table4,
    use_pragma_unroll,
)
from keopscore.mapreduce.Chunk_Mode_Constants import Chunk_Mode_Constants


def do_chunk_sub(
    dtype,
    red_formula,
    fun_chunked_curr,
    dimchunk_curr,
    dimsx,
    dimsy,
    indsi,
    indsj,
    indsi_chunked,
    indsj_chunked,
    acc,
    tile,
    i,
    j,
    jstart,
    chunk,
    nx,
    ny,
    arg,
    fout,
    xi,
    yj,
    yjrel,
    param_loc,
):
    chk = Chunk_Mode_Constants(red_formula)
    fout_tmp_chunk = c_array(dtype, chk.fun_chunked.dim)
    xiloc = c_variable(pointer(dtype), f"({xi.id} + {chk.dimx_notchunked})")
    yjloc = c_variable(
        pointer(dtype), f"({yj.id} + threadIdx.x * {chk.dimy} + {chk.dimy_notchunked})"
    )
    load_chunks_routine_i = load_vars_chunks(
        indsi_chunked,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        xiloc,
        arg,
        chunk,
        row_index=i,
    )
    load_chunks_routine_j = load_vars_chunks(
        indsj_chunked,
        dimchunk,
        dimchunk_curr,
        chk.dim_org,
        yjloc,
        arg,
        chunk,
        row_index=j,
    )
    chktable = table(
        chk.nminargs,
        dimsx,
        dimsy,
        chk.dimsp,
        indsi,
        indsj,
        chk.indsp,
        xi,
        yjrel,
        param_loc,
    )
    foutj = c_variable(pointer(dtype), "foutj")

    return f"""
                // Starting chunk_sub routine
                {fout_tmp_chunk.declare()}
                if ({i.id} < {nx.id}) {{
                    {load_chunks_routine_i}
                }} 
                if (j < ny) {{ // we load yj from device global memory only if j<ny
                    {load_chunks_routine_j}
                }}
                __syncthreads();
                if ({i.id} < {nx.id}) {{ // we compute only if needed
                    {dtype} *yjrel = {yj.id}; // Loop on the columns of the current block.
                    for (int jrel = 0; (jrel < blockDim.x) && (jrel < {ny.id} - jstart); jrel++, yjrel += {chk.dimy}) {{
                        {dtype} *foutj = {fout.id} + jrel*{chk.fun_chunked.dim};
                        {fun_chunked_curr(fout_tmp_chunk, chktable)}
                        {chk.fun_chunked.acc_chunk(foutj, fout_tmp_chunk)}
                    }}
                }} 
                __syncthreads();
                // Finished chunk_sub routine
            """


class GpuReduc1D_chunks(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.chk = Chunk_Mode_Constants(self.red_formula)
        self.dimy = self.chk.dimy
        self.blocksize_chunks = min(
            cuda_block_size, 1024, 49152 // max(1, self.dimy * sizeof(self.dtype))
        )

    def get_code(self):
        super().get_code()

        red_formula = self.red_formula
        dtype = self.dtype
        dtypeacc = self.dtypeacc
        varloader = self.varloader

        i = self.i
        j = self.j

        arg = self.arg
        args = self.args

        yjrel = c_array(dtype, varloader.dimy, "yjrel")

        jreltile = c_variable("int", "(jrel + tile * blockDim.x)")

        chk = self.chk
        param_loc = c_array(dtype, chk.dimp, "param_loc")
        acc = c_array(dtypeacc, chk.dimred, "acc")
        sum_scheme = eval(self.sum_scheme_string)(red_formula, dtype, dimred=chk.dimred)
        xi = c_array(dtype, chk.dimx, "xi")
        fout_chunk = c_array(
            dtype, self.blocksize_chunks * chk.dimout_chunk, "fout_chunk"
        )
        yj = c_variable(pointer(dtype), "yj")
        yjloc = c_array(dtype, chk.dimy, f"(yj + threadIdx.x * {chk.dimy})")

        fout_chunk_loc = c_variable(
            pointer(dtype), f"({fout_chunk.id}+jrel*{chk.dimout_chunk})"
        )

        tile = c_variable("int", "tile")
        nx = c_variable("int", "nx")
        ny = c_variable("int", "ny")

        jstart = c_variable("int", "jstart")
        chunk = c_variable("int", "chunk")

        chunk_sub_routine = do_chunk_sub(
            dtype,
            red_formula,
            chk.fun_chunked,
            dimchunk,
            chk.dimsx,
            chk.dimsy,
            chk.indsi,
            chk.indsj,
            chk.indsi_chunked,
            chk.indsj_chunked,
            acc,
            tile,
            i,
            j,
            jstart,
            chunk,
            nx,
            ny,
            arg,
            fout_chunk,
            xi,
            yj,
            yjrel,
            param_loc,
        )

        last_chunk = c_variable("int", f"{chk.nchunks - 1}")
        chunk_sub_routine_last = do_chunk_sub(
            dtype,
            red_formula,
            chk.fun_lastchunked,
            chk.dimlastchunk,
            chk.dimsx_last,
            chk.dimsy_last,
            chk.indsi,
            chk.indsj,
            chk.indsi_lastchunked,
            chk.indsj_lastchunked,
            acc,
            tile,
            i,
            j,
            jstart,
            last_chunk,
            nx,
            ny,
            arg,
            fout_chunk,
            xi,
            yj,
            yjrel,
            param_loc,
        )

        foutj = c_array(dtype, chk.dimout_chunk, "foutj")
        chktable_out = table4(
            chk.nminargs + 1,
            chk.dimsx,
            chk.dimsy,
            chk.dimsp,
            [chk.dimout_chunk],
            chk.indsi,
            chk.indsj,
            chk.indsp,
            [chk.nminargs],
            xi,
            yjrel,
            param_loc,
            foutj,
        )
        fout_tmp = c_array(dtype, chk.dimfout, "fout_tmp")
        outi = c_array(dtype, chk.dimout, f"(out + i * {chk.dimout})")

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void GpuConv1DOnDevice(int nx, int ny, {dtype} *out, {dtype} **{arg.id}) {{
    
                          // get the index of the current thread
                          int i = blockIdx.x * blockDim.x + threadIdx.x;

                          // declare shared mem
                          extern __shared__ {dtype} yj[];

                          // load parameters variables from global memory to local thread memory
                          {param_loc.declare()}
                          {load_vars(chk.dimsp, chk.indsp, param_loc, args)}
                          
                          {acc.declare()}
                          
                          {sum_scheme.declare_temporary_accumulator()}                     
                          
                          if (i < nx) {{
                            {red_formula.InitializeReduction(acc)} // acc = 0
                            {sum_scheme.initialize_temporary_accumulator_first_init()}
                          }}

                          {xi.declare()}

                          {fout_chunk.declare()}
                          
                          if (i < nx) {{
                            {load_vars(chk.dimsx_notchunked, chk.indsi_notchunked, xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                          }}

                          for (int jstart = 0, tile = 0; jstart < ny; jstart += blockDim.x, tile++) {{

                            // get the current column
                            int j = tile * blockDim.x + threadIdx.x;

                            if (j < ny) {{ // we load yj from device global memory only if j<ny
                              {load_vars(chk.dimsy_notchunked, chk.indsj_notchunked, yjloc, args, row_index=j)} 
                            }}
                            __syncthreads();

                            if (i < nx) {{ // we compute x1i only if needed
                              for (int jrel = 0; (jrel < blockDim.x) && (jrel < ny - jstart); jrel++) {{
                                {chk.fun_chunked.initacc_chunk(fout_chunk_loc)}
                              }}
                              {sum_scheme.initialize_temporary_accumulator_block_init()}
                            }}
                            
                            // looping on chunks (except the last)
                    		{use_pragma_unroll()}
                    		for (int chunk=0; chunk<{chk.nchunks}-1; chunk++) {{
                              {chunk_sub_routine}
                            }}
                            // last chunk
                            {chunk_sub_routine_last}
                            
                            if (i < nx) {{
                                {dtype} * yjrel = yj; // Loop on the columns of the current block.
                                for (int jrel = 0; (jrel < blockDim.x) && (jrel < ny - jstart); jrel++, yjrel += {chk.dimy}) {{
                                    {dtype} *foutj = fout_chunk + jrel*{chk.dimout_chunk};
                                    {fout_tmp.declare()}
                                    {chk.fun_postchunk(fout_tmp, chktable_out)}
                                    {sum_scheme.accumulate_result(acc, fout_tmp, jreltile)}
                                }}
                                {sum_scheme.final_operation(acc)}
                            }}
                            __syncthreads();
                          }}

                          if (i < nx) {{
                            {red_formula.FinalizeOutput(acc, outi, i)} 
                          }}
                        }}
                    """



from keopscore.formulas.maths.SqNorm2 import SqNorm2


class SqNormIso:
    """
    ISOTROPIC NORM : SqNormIso(S,A)
    """

    def __new__(cls, S, A):
        return S * SqNorm2(A)

    enable_test = False



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.math_functions import keops_round


class Round(VectorizedScalarOp):
    """the Round vectorized operation
    Round(f,d) where d is integer, rounds f to d decimals
    """

    def __init__(self, f, d):
        super().__init__(f, params=(d,))

    string_id = "Round"

    ScalarOpFun = keops_round

    def DiffT(self, v, gradin):
        return Zero(v.dim)

    # parameters for testing the operation (optional)
    nargs = 1  # number of arguments
    test_params = [3]  # parameters to try
    torch_op = None  # equivalent PyTorch operation



from keopscore.formulas.reductions import *
from keopscore.formulas.GetReduction import GetReduction
from keopscore.utils.code_gen_utils import Var_loader, new_c_varname, pointer, c_include


class MapReduce:
    """
    base class for map-reduce schemes
    """

    def __init__(
        self,
        red_formula_string,
        aliases,
        nargs,
        dtype,
        dtypeacc,
        sum_scheme_string,
        tagHostDevice,
        tagCpuGpu,
        tag1D2D,
        use_half,
        device_id,
    ):
        self.red_formula_string = red_formula_string
        self.aliases = aliases

        self.red_formula = GetReduction(red_formula_string, aliases=aliases)

        self.dtype = dtype
        self.dtypeacc = dtypeacc
        self.nargs = nargs
        self.sum_scheme_string = sum_scheme_string
        self.tagHostDevice, self.tagCpuGpu, self.tag1D2D = (
            tagHostDevice,
            tagCpuGpu,
            tag1D2D,
        )
        self.use_half = use_half
        self.device_id = device_id
        self.varloader = Var_loader(self.red_formula)

    def get_code(self):

        self.headers = "#define C_CONTIGUOUS 1\n"

        if self.use_half == 1:
            self.headers += "#define USE_HALF 1\n"
            self.headers += c_include("cuda_fp16.h")
        else:
            self.headers += "#define USE_HALF 0\n"

        red_formula = self.red_formula
        formula = red_formula.formula
        dtype = self.dtype
        dtypeacc = self.dtypeacc
        nargs = self.nargs
        self.sum_scheme = eval(self.sum_scheme_string)(red_formula, dtype)

        self.i = i = c_variable("int", "i")
        self.j = j = c_variable("int", "j")

        nx = c_variable("int", "nx")
        ny = c_variable("int", "ny")

        self.xi = c_array(dtype, self.varloader.dimx, "xi")
        self.param_loc = c_array(dtype, self.varloader.dimp, "param_loc")

        argname = new_c_varname("arg")
        self.arg = c_variable(pointer(pointer(dtype)), argname)
        self.args = [self.arg[k] for k in range(nargs)]

        self.acc = c_array(dtypeacc, red_formula.dimred, "acc")
        self.acctmp = c_array(dtypeacc, red_formula.dimred, "acctmp")
        self.fout = c_array(dtype, formula.dim, "fout")
        self.outi = c_array(dtype, red_formula.dim, f"(out + i * {red_formula.dim})")



from keopscore.utils.code_gen_utils import ComplexVectApply
from keopscore.formulas.Operation import Operation
from keopscore.utils.misc_utils import KeOps_Error


class VectorizedComplexScalarOp(Operation):
    # class for operations that are vectorized or broadcasted
    # complex scalar operations,
    # such as ComplexExp(f), ComplexMult(f), ComplexAdd(f,g), etc.

    def __init__(self, *args, params=()):
        dims = set(arg.dim for arg in args)
        if max(dims) % 2 != 0 or len(dims) > 2 or (len(dims) == 2 and min(dims) != 2):
            KeOps_Error("dimensions are not compatible for VectorizedComplexScalarOp")
        super().__init__(*args, params=params)

    @property
    def dim(self):
        # dim gives the output dimension of the operation,
        # here it is the same as the output dimension of the child operation
        return max(child.dim for child in self.children)

    def Op(self, out, table, *args):
        # Atomic evaluation of the operation : it consists in a simple
        # for loop around the call to the correponding scalar operation
        return ComplexVectApply(self.ScalarOp, out, *args)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.maths.ClampInt import ClampInt
from keopscore.utils.math_functions import keops_clamp


class Clamp(VectorizedScalarOp):
    """Clamp(x,a,b) = a if x<a, x if a<=x<=b, b if b<x"""

    string_id = "Clamp"

    ScalarOpFun = keops_clamp

    def DiffT(self, v, gradin):
        # N.B.   Clamp(F,G,H) = Clamp((F-G)/(H-G),0,1) * (H-G) + G
        # We use this fact to avoid writing another custom operation for the gradient.
        # (This may be slower however...)
        f, g, h = self.children
        Alt_Clamp = ClampInt((f - g) / (h - g), 0, 1) * (h - g) + g
        return Alt_Clamp.DiffT(v, gradin)

    # parameters for testing the operation (optional)
    nargs = 3  # number of arguments
    torch_op = None  # equivalent PyTorch operation



import os
from ctypes import create_string_buffer, CDLL, c_int
from os import RTLD_LAZY
import sysconfig

from keopscore.binders.LinkCompile import LinkCompile
import keopscore.config
from keopscore.config.config import (
    cuda_version,
    jit_binary,
    cxx_compiler,
    nvrtc_flags,
    nvrtc_include,
    jit_source_file,
    cuda_available,
    get_build_folder,
)
from keopscore.utils.misc_utils import KeOps_Error, KeOps_Message, KeOps_OS_Run
from keopscore.utils.gpu_utils import get_gpu_props, cuda_include_fp16_path

jit_compile_src = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "nvrtc_jit.cpp"
)


def jit_compile_dll():
    return os.path.join(
        get_build_folder(),
        "nvrtc_jit" + sysconfig.get_config_var("SHLIB_SUFFIX"),
    )


class Gpu_link_compile(LinkCompile):
    source_code_extension = "cu"
    low_level_code_prefix = "cubin_" if cuda_version >= 11010 else "ptx_"
    ngpu, gpu_props_compile_flags = get_gpu_props()

    def __init__(self):
        # checking that the system has a Gpu :
        if not (cuda_available and Gpu_link_compile.ngpu):
            KeOps_Error(
                "Trying to compile cuda code... but we detected that the system has no properly configured cuda lib."
            )

        LinkCompile.__init__(self)
        # these are used for JIT compiling mode
        # low_level_code_file is filename of low level code (PTX for Cuda) or binary (CUBIN for Cuda)
        # generated by the JIT compiler, e.g. ptx_7b9a611f7e
        self.low_level_code_file = os.path.join(
            get_build_folder(),
            self.low_level_code_prefix + self.gencode_filename,
        ).encode("utf-8")

        self.my_c_dll = CDLL(jit_compile_dll(), mode=RTLD_LAZY)
        # actual dll to be called is the jit binary, TODO: check if this is relevent
        self.true_dllname = jit_binary
        # file to check for existence to detect compilation is needed
        self.file_to_check = self.low_level_code_file

    def generate_code(self):
        # method to generate the code and compile it
        # generate the code and save it in self.code, by calling get_code method from GpuReduc class :
        self.get_code()
        # write the code in the source file
        self.write_code()
        # we execute the main dll, passing the code as argument, and the name of the low level code file to save the assembly instructions
        self.my_c_dll.Compile(
            create_string_buffer(self.low_level_code_file),
            create_string_buffer(self.code.encode("utf-8")),
            c_int(self.use_half),
            c_int(self.device_id),
            create_string_buffer(
                (cuda_include_fp16_path() + os.path.sep).encode("utf-8")
            ),
        )
        # retreive some parameters that will be saved into info_file.
        self.tagI = self.red_formula.tagI
        self.dim = self.red_formula.dim

    @staticmethod
    def get_compile_command(
        sourcename=jit_source_file, dllname=jit_binary, extra_flags=""
    ):
        # This is about the main KeOps binary (dll) that will be used to JIT compile all formulas.
        # If the dll is not present, it compiles it from source, except if check_compile is False.

        target_tag = (
            "CUBIN" if Gpu_link_compile.low_level_code_prefix == "cubin_" else "PTX"
        )
        nvrtcGetTARGET = "nvrtcGet" + target_tag
        nvrtcGetTARGETSize = nvrtcGetTARGET + "Size"
        arch_tag = (
            '\\"sm\\"'
            if Gpu_link_compile.low_level_code_prefix == "cubin_"
            else '\\"compute\\"'
        )
        target_type_define = f"-DnvrtcGetTARGET={nvrtcGetTARGET} -DnvrtcGetTARGETSize={nvrtcGetTARGETSize} -DARCHTAG={arch_tag}"
        return f"{cxx_compiler} {nvrtc_flags} {extra_flags} {target_type_define} {nvrtc_include} {Gpu_link_compile.gpu_props_compile_flags} {sourcename} -o {dllname}"

    @staticmethod
    def compile_jit_compile_dll():
        KeOps_Message("Compiling cuda jit compiler engine ... ", flush=True, end="")
        KeOps_OS_Run(
            Gpu_link_compile.get_compile_command(
                sourcename=jit_compile_src,
                dllname=jit_compile_dll(),
            ),
        )
        KeOps_Message("OK", use_tag=False, flush=True)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.utils.math_functions import keops_exp


##########################
######    Exp        #####
##########################


class Exp(VectorizedScalarOp):
    """the exponential vectorized operation"""

    string_id = "Exp"

    ScalarOpFun = keops_exp

    @staticmethod
    def Derivative(f):
        return Exp(f)



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_for_loop
from keopscore.formulas.complex.ComplexReal import ComplexReal
from keopscore.formulas.complex.ComplexMult import ComplexMult
from keopscore.formulas.complex.Conj import Conj
from keopscore.utils.misc_utils import KeOps_Error

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexSquareAbs                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexSquareAbs(Operation):

    string_id = "ComplexSquareAbs"

    def __init__(self, f):
        if f.dim % 2 != 0:
            KeOps_Error("Dimension of F must be even")
        self.dim = int(f.dim / 2)
        super().__init__(f)

    def Op(self, out, table, inF):
        f = self.children[0]
        forloop, i = c_for_loop(0, f.dim, 2, pragma_unroll=True)
        return forloop(out[i / 2].assign(inF[i] * inF[i] + inF[i + 1] * inF[i + 1]))

    def DiffT(self, v, gradin):
        f = self.children[0]
        AltFormula = ComplexReal(ComplexMult(f, Conj(f)))
        return AltFormula.DiffT(v, gradin)



from keopscore.formulas.VectorizedComplexScalarOp import VectorizedComplexScalarOp

# /////////////////////////////////////////////////////////////////////////
# ////      ComplexSubtract                           ////
# /////////////////////////////////////////////////////////////////////////


class ComplexSubtract(VectorizedComplexScalarOp):

    string_id = "ComplexSubtract"

    def ScalarOp(self, out, inF, inG):
        string = out[0].assign(inF[0] - inG[0])
        string += out[1].assign(inF[1] - inG[1])
        return string

    def DiffT(self, v, gradin):
        f, g = self.children
        return ComplexSubtract(f.DiffT(v, gradin), g.DiffT(v, gradin))



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp
from keopscore.formulas.variables.Zero import Zero
from keopscore.utils.math_functions import keops_step


class Step(VectorizedScalarOp):
    """the Step vectorized operation"""

    string_id = "Step"

    ScalarOpFun = keops_step

    def DiffT(self, v, gradin):
        return Zero(v.dim)



import os
from os.path import join
import shutil
from ctypes import CDLL, RTLD_GLOBAL
import keopscore
from ctypes.util import find_library
from keopscore.utils.misc_utils import KeOps_Warning, KeOps_Error
import platform, sys

# global parameters can be set here :
use_cuda = True  # use cuda if possible
use_OpenMP = True  # use OpenMP if possible (see function set_OpenMP below)

# System Path
base_dir_path = os.path.abspath(join(os.path.dirname(os.path.realpath(__file__)), ".."))
template_path = join(base_dir_path, "templates")
bindings_source_dir = join(base_dir_path)
keops_cache_folder = join(
    os.path.expanduser("~"), ".cache", f"keops{keopscore.__version__}"
)
default_build_folder_name = "build"
# In case user has specified CUDA_VISIBLE_DEVICES environment variable,
# it is better to set the build folder name accordingly.
specific_gpus = os.getenv("CUDA_VISIBLE_DEVICES")
if specific_gpus:
    specific_gpus = specific_gpus.replace(",", "_")
    default_build_folder_name += "_CUDA_VISIBLE_DEVICES_" + specific_gpus
default_build_path = join(keops_cache_folder, default_build_folder_name)

# init cache folder
os.makedirs(keops_cache_folder, exist_ok=True)


# build path setter/getter

_build_path = None


def set_build_folder(
    path=None, read_save_file=False, write_save_file=True, reset_all=True
):

    # if path is not given, we either read the save file or use the default build path
    save_file = join(keops_cache_folder, "build_folder_location.txt")
    if not path:
        if read_save_file and os.path.isfile(save_file):
            f = open(save_file, "r")
            path = f.read()
            f.close()
        else:
            path = default_build_path

    # create the folder if not yet done
    os.makedirs(path, exist_ok=True)

    # _build_path contains the current build folder path (or None if not yet set). We need
    # to remove this _build_path from the sys.path, replace the value of _build_path
    # and update the sys.path
    global _build_path
    if _build_path in sys.path:
        sys.path.remove(_build_path)
    _build_path = path
    sys.path.append(path)

    # saving the location of the build path in a file
    if write_save_file:
        f = open(save_file, "w")
        f.write(path)
        f.close()

    # reset all cached formulas if needed
    if reset_all:
        keopscore.get_keops_dll.get_keops_dll.reset(new_save_folder=_build_path)
        if keopscore.config.config.use_cuda:
            from keopscore.binders.nvrtc.Gpu_link_compile import (
                Gpu_link_compile,
                jit_compile_dll,
            )

            if not os.path.exists(jit_compile_dll()):
                Gpu_link_compile.compile_jit_compile_dll()


set_build_folder(read_save_file=True, write_save_file=False, reset_all=False)


def get_build_folder():
    return _build_path


jit_binary = join(_build_path, "keops_nvrtc.so")

# Compiler
cxx_compiler = os.getenv("CXX")
if cxx_compiler is None:
    cxx_compiler = "g++"
if shutil.which(cxx_compiler) is None:
    KeOps_Warning(
        """
    The default C++ compiler could not be found on your system.
    You need to either define the CXX environment variable or a symlink to the g++ command.
    For example if g++-8 is the command you can do
      import os
      os.environ['CXX'] = 'g++-8'
    """
    )


compile_options = " -shared -fPIC -O3 -std=c++11"


# cpp options
if platform.system() == "Darwin":
    cpp_flags = compile_options + " -flto"
else:
    cpp_flags = compile_options + " -flto=auto"

disable_pragma_unrolls = True

# OpenMP setting
# adds compile flags for OpenMP support.
if use_OpenMP:
    if platform.system() == "Darwin":
        import subprocess, importlib

        res = subprocess.run(
            'echo "#include <omp.h>" | g++ -E - -o /dev/null',
            stdout=subprocess.PIPE,
            shell=True,
        )
        if res.returncode != 0:
            KeOps_Warning("omp.h header is not in the path, disabling OpenMP.")
            use_OpenMP = False
        else:
            # we try to import either mkl or numpy, because it will load
            # the shared libraries for OpenMP.
            import importlib.util

            if importlib.util.find_spec("mkl"):
                import mkl
            elif importlib.util.find_spec("numpy"):
                import numpy
            # Now we can look if one of libmkl_rt, libomp and/or libiomp is loaded.
            pid = os.getpid()
            loaded_libs = {}
            for lib in ["libomp", "libiomp", "libmkl_rt"]:
                res = subprocess.run(
                    f"lsof -p {pid} | grep {lib}", stdout=subprocess.PIPE, shell=True
                )
                loaded_libs[lib] = (
                    os.path.dirname(res.stdout.split(b" ")[-1]).decode("utf-8")
                    if res.returncode == 0
                    else None
                )
            if loaded_libs["libmkl_rt"]:
                cpp_flags += f' -Xclang -fopenmp -lmkl_rt -L{loaded_libs["libmkl_rt"]}'
            elif loaded_libs["libiomp"]:
                cpp_flags += f' -Xclang -fopenmp -liomp5 -L{loaded_libs["libiomp"]}'
            elif loaded_libs["libomp"]:
                cpp_flags += f' -Xclang -fopenmp -lomp -L{loaded_libs["libomp"]}'
            else:
                KeOps_Warning("OpenMP shared libraries not loaded, disabling OpenMP.")
                use_OpenMP = False
    else:
        cpp_flags += " -fopenmp -fno-fat-lto-objects"

if platform.system() == "Darwin":
    cpp_flags += " -undefined dynamic_lookup"

cpp_flags += " -I" + bindings_source_dir


from keopscore.utils.gpu_utils import get_gpu_props

cuda_dependencies = ["cuda", "nvrtc"]
if all([find_library(lib) for lib in cuda_dependencies]):
    # N.B. calling get_gpu_props issues a warning if cuda is not available, so we do not add another warning here
    cuda_available = get_gpu_props()[0] > 0
else:
    cuda_available = False
    KeOps_Warning(
        "Cuda libraries were not detected on the system ; using cpu only mode"
    )

if not use_cuda and cuda_available:
    KeOps_Warning(
        "Cuda appears to be available on your system, but use_cuda is set to False in config.py. Using cpu only mode"
    )

if use_cuda and not cuda_available:
    use_cuda = False

if use_cuda:
    from keopscore.utils.gpu_utils import (
        libcuda_folder,
        libnvrtc_folder,
        get_cuda_include_path,
        get_cuda_version,
    )

    cuda_version = get_cuda_version()
    nvrtc_flags = (
        compile_options
        + f" -fpermissive -L{libcuda_folder} -L{libnvrtc_folder} -lcuda -lnvrtc"
    )
    nvrtc_include = " -I" + bindings_source_dir
    cuda_include_path = get_cuda_include_path()
    if cuda_include_path:
        nvrtc_include += " -I" + cuda_include_path
    jit_source_file = join(base_dir_path, "binders", "nvrtc", "keops_nvrtc.cpp")
    jit_source_header = join(base_dir_path, "binders", "nvrtc", "keops_nvrtc.h")
else:
    cuda_version = None
    libcuda_folder = None
    libnvrtc_folder = None
    nvrtc_flags = None
    nvrtc_include = None
    cuda_include_path = None
    jit_source_file = None
    jit_source_header = None
    jit_binary = None

init_cudalibs_flag = False


def init_cudalibs():
    if not keopscore.config.config.init_cudalibs_flag:
        # we load some libraries that need to be linked with KeOps code
        # This is to avoid "undefined symbols" errors.
        CDLL(find_library("nvrtc"), mode=RTLD_GLOBAL)
        CDLL(find_library("cuda"), mode=RTLD_GLOBAL)
        CDLL(find_library("cudart"), mode=RTLD_GLOBAL)
        keopscore.config.config.init_cudalibs_flag = True



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_array, VectCopy
from keopscore.utils.misc_utils import KeOps_Error

# //////////////////////////////////////////////////////////////
# ////     VECTOR EXTRACTION : Extract<F,START,DIM>         ////
# //////////////////////////////////////////////////////////////


class Extract(Operation):
    string_id = "Extract"

    def __init__(self, arg0, start, dim):
        if arg0.dim < start + dim or start < 0:
            KeOps_Error("Index out of bound in Extract")
        super().__init__(arg0, params=(start, dim))
        self.start = start
        self.dim = dim

    def Op(self, out, table, arg0):
        # returns the atomic piece of c++ code to evaluate the function on arg and return
        # the result in out
        v = c_array(arg0.dtype, out.dim, f"({arg0.id}+{self.start})")
        return VectCopy(out, v)

    def DiffT(self, v, gradin):
        from keopscore.formulas.maths.ExtractT import ExtractT

        f = self.children[0]
        return f.DiffT(v, ExtractT(gradin, self.start, f.dim))

    # parameters for testing the operation (optional)
    enable_test = True  # enable testing for this operation
    nargs = 1  # number of arguments
    test_argdims = [10]  # dimensions of arguments for testing
    test_params = [3, 5]  # values of parameters for testing
    torch_op = "lambda x,s,d : x[...,s:(s+d)]"  # equivalent PyTorch operation



import ast, inspect

import keopscore.formulas
from keopscore.utils.code_gen_utils import get_hash_name
from keopscore.formulas.reductions import *
from keopscore.formulas.maths import *
from keopscore.formulas.complex import *
from keopscore.formulas.variables import *
from keopscore.formulas.autodiff import *


class GetReduction:
    library = {}

    def __new__(self, red_formula_string, aliases=[]):
        string_id_hash = get_hash_name(red_formula_string, aliases)
        if string_id_hash in GetReduction.library:
            return GetReduction.library[string_id_hash]
        else:
            self.check_formula(red_formula_string)
            aliases_dict = {}
            for alias in aliases:
                self.check_formula(alias)
                if "=" in alias:
                    varname, var = alias.split("=")
                    aliases_dict[varname] = eval(var)
            reduction = eval(red_formula_string, globals(), aliases_dict)
            GetReduction.library[string_id_hash] = reduction
            return reduction

    @staticmethod
    def check_formula(string):
        formula_class = dict(inspect.getmembers(keopscore.formulas))
        parsed = ast.parse(string)
        for node in ast.walk(parsed):
            if isinstance(node, ast.Call) and (
                node.func.id not in formula_class.keys()
            ):
                print(node.func.id)
                # raise NotImplementedError



from keopscore.formulas.Operation import Operation
from keopscore.utils.code_gen_utils import c_zero_float, c_for_loop

# /////////////////////////////////////////////////////////////////////////
# ////      Real2Complex                           ////
# /////////////////////////////////////////////////////////////////////////


class Real2Complex(Operation):

    string_id = "Real2Complex"

    def __init__(self, f):
        self.dim = 2 * f.dim
        super().__init__(f)

    def Op(self, out, table, inF):
        forloop, i = c_for_loop(0, self.dim, 2, pragma_unroll=True)
        body = out[i].assign(inF[i / 2])
        body += out[i + 1].assign(c_zero_float)
        return forloop(body)

    def DiffT(self, v, gradin):
        from keopscore.formulas.complex.ComplexReal import ComplexReal

        f = self.children[0]
        return f.DiffT(v, ComplexReal(gradin))



# special computation scheme for dim>100

enable_chunk = True


def get_enable_chunk():
    global enable_chunk
    return enable_chunk


def set_enable_chunk(val):
    global enable_chunk
    if val == 1:
        enable_chunk = True
    elif val == 0:
        enable_chunk = False


dimchunk = 64
dim_treshold_chunk = 146
specdims_use_chunk = [99, 100, 102, 120, 133, 138, 139, 140, 141, 142]

# special mode for formula of the type sum_j k(x_i,y_j)*b_j with high dimensional b_j
enable_final_chunk = True


def set_enable_finalchunk(val):
    global enable_final_chunk
    if val == 1:
        enable_final_chunk = True
    elif val == 0:
        enable_final_chunk = False


dimfinalchunk = 64


def get_dimfinalchunk():
    global dimfinalchunk
    return dimfinalchunk


def set_dimfinalchunk(val):
    global dimfinalchunk
    dimfinalchunk = val


def use_final_chunks(red_formula):
    global enable_final_chunk
    global mult_var_highdim
    global dim_treshold_chunk
    return (
        enable_final_chunk and mult_var_highdim and red_formula.dim > dim_treshold_chunk
    )


mult_var_highdim = False


def set_mult_var_highdim(val):
    global mult_var_highdim
    if val == 1:
        mult_var_highdim = True
    elif val == 0:
        mult_var_highdim = False



# same as Grad with additional saved forward variable. This is only used for taking gradients of reductions operations.


def Grad_WithSavedForward(red_formula, v, gradin, f0):
    return red_formula.DiffT(v, gradin, f0)



from keopscore.utils.code_gen_utils import VectApply, VectCopy
from keopscore.utils.Tree import Tree


class Reduction(Tree):
    """Base class for all KeOps final reductions over a formula"""

    def __init__(self, formula, tagI):
        """- formula is an object of type Operation, it is the formula on which we apply a reduction
        - tagI : 0 or 1, specifies wether we do the reduction over "i"-indexed or "j"-indexed variables."""

        # We initialize several constants, most of them infered from the formula
        self.formula = formula
        self.children = [formula]
        self.params = (tagI,)
        self.tagI = tagI
        self.tagJ = 1 - tagI
        self.cat = tagI
        self.Vars_ = formula.Vars_

    def ReducePair(self, acc, xi):
        """Returns C++ code that implements the update phase of the reduction.
        by default it consists in a vectorized version of the ReducePairScalar operation."""
        return VectApply(self.ReducePairScalar, acc, xi)

    def ReducePairShort(self, acc, xi, ind):
        # N.B next lines are useless here, but to be used in other reductions :
        # if xi.dtype == "half2":
        #     half2_val = c_variable("half2_ind")
        #     string = half2_val.declare_assign(f"__floats2half2_rn(2*{ind()},2*{ind()}+1)")
        return self.ReducePair(acc, xi)

    def FinalizeOutput(self, acc, out, i):
        """Returns C++ code that implements the final output of the reduction.
        For most reducitons it is a simple copy of the temporary variable
        updated during the reduction, with possibly a cast if the accumulator was of
        different data type."""
        return VectCopy(out, acc)



from keopscore.formulas.VectorizedScalarOp import VectorizedScalarOp

from keopscore.formulas.variables.Zero import Zero


##########################
######    Square     #####
##########################


class Square_Impl(VectorizedScalarOp):
    """the square vectorized operation"""

    string_id = "Square"
    print_spec = "**2", "post", 1

    def ScalarOp(self, out, arg):
        return out.assign(arg * arg)

    @staticmethod
    def Derivative(f):
        return 2 * f


# N.B. The following separate function should theoretically be implemented
# as a __new__ method of the previous class, but this can generate infinite recursion problems
def Square(arg):
    if isinstance(arg, Zero):
        return arg
    else:
        return Square_Impl(arg)



import os

import keopscore.config.config
from keopscore.config.config import get_build_folder
from keopscore.utils.code_gen_utils import get_hash_name
from keopscore.utils.misc_utils import KeOps_Error, KeOps_Message
from keopscore.config.config import cpp_flags


class LinkCompile:
    """
    Base class for compiling the map_reduce schemes and providing the dll to KeOps bindings.
    """

    def __init__(self):
        # N.B. Here self is assumed to be populated by the __init__ of one of the MapReduce classes

        # we create the hash string id corresponding to all parameters, e.g. 7b9a611f7e
        self.gencode_filename = get_hash_name(
            type(self),
            self.red_formula_string,
            self.aliases,
            self.nargs,
            self.dtype,
            self.dtypeacc,
            self.sum_scheme_string,
            self.tagHostDevice,
            self.tagCpuGpu,
            self.tag1D2D,
            self.use_half,
            self.device_id,
            cpp_flags,
        )

        # info_file is the name of the file that will contain some meta-information required by the bindings, e.g. 7b9a611f7e.nfo
        self.info_file = os.path.join(
            get_build_folder(), self.gencode_filename + ".nfo"
        )

        # gencode_file is the name of the source file to be created and then compiled, e.g. 7b9a611f7e.cpp or 7b9a611f7e.cu
        self.gencode_file = os.path.join(
            get_build_folder(),
            self.gencode_filename + "." + self.source_code_extension,
        )

    def save_info(self):
        # create info_file to save some parameters : dim (dimension of output vectors),
        #                                            tagI (O or 1, reduction over i or j indices),
        #                                            dimy (sum of dimensions of j-indexed vectors)
        f = open(self.info_file, "w")
        f.write(
            f"red_formula={self.red_formula_string}\ndim={self.dim}\ntagI={self.tagI}\ndimy={self.dimy}"
        )
        f.close()

    def read_info(self):
        # read info_file to retreive dim, tagI, dimy
        f = open(self.info_file, "r")
        string = f.read()
        f.close()
        tmp = string.split("\n")
        if len(tmp) != 4:
            KeOps_Error("Incorrect info file")
        tmp_dim, tmp_tag, tmp_dimy = (
            tmp[1].split("="),
            tmp[2].split("="),
            tmp[3].split("="),
        )
        if (
            len(tmp_dim) != 2
            or tmp_dim[0] != "dim"
            or len(tmp_tag) != 2
            or tmp_tag[0] != "tagI"
            or len(tmp_dimy) != 2
            or tmp_dimy[0] != "dimy"
        ):
            KeOps_Error("Incorrect info file")
        self.dim = eval(tmp_dim[1])
        self.tagI = eval(tmp_tag[1])
        self.dimy = eval(tmp_dimy[1])

    def write_code(self):
        # write the generated code in the source file ; this is used as a subfunction of compile_code
        f = open(self.gencode_file, "w")
        f.write(self.code)
        f.close()

    def generate_code(self):
        pass

    def get_dll_and_params(self):
        # main method of the class : it generates - if needed - the code and returns the name of the dll to be run for
        # performing the reduction, e.g. 7b9a611f7e.so, or in the case of JIT compilation, the name of the main KeOps dll,
        # and the name of the assembly code file.
        if not os.path.exists(self.file_to_check):
            KeOps_Message(
                "Generating code for formula " + self.red_formula.__str__() + " ... ",
                flush=True,
                end="",
            )
            self.generate_code()
            self.save_info()
            KeOps_Message("OK", use_tag=False, flush=True)
        else:
            self.read_info()
        return dict(
            tag=self.gencode_filename,
            source_file=self.true_dllname,
            low_level_code_file=self.low_level_code_file,
            tagI=self.tagI,
            use_half=self.use_half,
            tag1D2D=self.tag1D2D,
            dimred=self.red_formula.dimred,
            dim=self.dim,
            dimy=self.dimy,
            indsi=self.varloader.indsi,
            indsj=self.varloader.indsj,
            indsp=self.varloader.indsp,
            dimsx=self.varloader.dimsx,
            dimsy=self.varloader.dimsy,
            dimsp=self.varloader.dimsp,
        )



from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.formulas.reductions.sum_schemes import block_sum, kahan_scheme
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import c_variable, c_array, use_pragma_unroll
from keopscore.utils.misc_utils import KeOps_Error


class GpuReduc2D(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        i = self.i
        j = self.j
        red_formula = self.red_formula
        dtype = self.dtype
        dimin = red_formula.dimred
        dimout = red_formula.dim
        varloader = self.varloader
        dtypeacc = self.dtypeacc
        acc2 = c_array(dtypeacc, dimin, "acc2")

        inloc = c_array(dtype, dimin, f"(in + (tid+y*nx)*{dimin})")
        outloc = c_array(dtype, dimout, f"(out+tid*{dimout})")

        dimsx = varloader.dimsx
        dimsy = varloader.dimsy
        dimsp = varloader.dimsp
        indsi = varloader.indsi
        indsj = varloader.indsj
        indsp = varloader.indsp
        dimx = sum(dimsx)
        dimy = sum(dimsy)
        dimp = sum(dimsp)
        dimred = red_formula.dimred
        dimfout = red_formula.formula.dim

        fout = c_array(dtype, dimfout, "fout")
        param_loc = c_array(dtype, dimp, "param_loc")
        xi = c_array(dtype, dimx, "xi")

        sum_scheme = self.sum_scheme

        # N.B. To be consistent with the convention used in GpuConv1D, when SUM_SCHEME == BLOCK_SUM=1 we accumulate results in TYPE
        # instead of __TYPEACC__ in each block, __TYPEACC__ will be used only to sum up results from each block
        if isinstance(sum_scheme, block_sum):
            acc = c_array(dtype, dimred, "acc")
        elif isinstance(sum_scheme, direct_sum):
            acc = c_array(dtypeacc, dimred, "acc")
        else:
            KeOps_Error("incorrect reduction scheme")

        yjloc = c_array(dtype, varloader.dimy, f"(yj + threadIdx.x * {varloader.dimy})")
        arg = self.arg
        args = self.args
        yjrel = c_array(dtype, dimy, "yjrel")
        table = varloader.table(self.xi, yjrel, self.param_loc)

        jrelloc = c_variable("int", "(blockDim.x*blockIdx.y+jrel)")

        tid = c_variable("int", "tid")

        self.code = f"""
                          
                        {self.headers}
                        
                        extern "C" __global__ void reduce2D({dtype} *in, {dtype} *out, int sizeY, int nx) {{
                            /* Function used as a final reduction pass in the 2D scheme,
                             * once the block reductions have been made.
                             * Takes as input:
                             * - in,  a  sizeY * (nx * DIMIN ) array
                             * - out, an          nx * DIMOUT   array
                             *
                             * Computes, in parallel, the "columnwise"-reduction (which correspond to lines of blocks)
                             * of *in and stores the result in out.
                             */
                            int tid = blockIdx.x * blockDim.x + threadIdx.x;

                            /* As shown below, the code that is used to store the block-wise sum
                              "tmp" in parallel is:
                                if(i<nx)
                                    for(int k=0; k<DIMX1; k++)
                                        (*px)[blockIdx.y*DIMX1*nx+i*DIMX1+k] = tmp[k];
                            */

                            /* // This code should be a bit more efficient (more parallel) in the case
                               // of a simple "fully parallel" reduction op such as "sum", "max" or "min"
                            TYPE res = 0;
                            if(tid < nx*DIMVECT) {{
                                for (int i = 0; i < sizeY; i++)
                                    res += in[tid + i*nx*DIMVECT]; // We use "+=" as a reduction op. But it could be anything, really!
                                // res = in[tid+ nx* DIMVECT];
                                out[tid] = res;
                            }}
                            */

                            // However, for now, we use a "vectorized" reduction op.,
                            // which can also handle non-trivial reductions such as "LogSumExp"
                            {acc2.declare()}
                            {red_formula.InitializeReduction(acc2)} // acc = 0
                            if(tid < nx) {{
                                for (int y = 0; y < sizeY; y++) {{
                                    {red_formula.ReducePair(acc2, inloc)} // acc += in[(tid+y*nx) *DIMVECT : +DIMVECT]; 
                                }}
                                {red_formula.FinalizeOutput(acc2, outloc, tid)}
                            }}
                        }}
                        
                        
                        
                        
                        extern "C" __global__ void GpuConv2DOnDevice(int nx, int ny, {dtype} *out, {dtype} **{arg.id}) {{
                            
                            {fout.declare()}
                            
                            // Load the parameter vector in the Thread Memory, for improved efficiency
                            {param_loc.declare()}
                            {varloader.load_vars("p", param_loc, args)} // load parameters variables from global memory to local thread memory
                            
                            // Weird syntax to create a pointer in shared memory.
                            extern __shared__ char yj_char[];
                            {dtype}* const yj = reinterpret_cast<{dtype}*>(yj_char);
                            
                            // Step 1 : Load in Thread Memory the information needed in the current line ---------------------------
                            int i = blockIdx.x * blockDim.x + threadIdx.x;
                            {xi.declare()}
                            
                            {acc.declare()}
                            
                            {sum_scheme.tmp_acc.declare()} //# if isinstance(sum_scheme, kahan_scheme) else ""
                            
                            if(i<nx) {{ // we will compute outi only if i is in the range
                                {red_formula.InitializeReduction(acc)} // acc = 0
                                {sum_scheme.tmp_acc.assign(c_zero_float) if isinstance(sum_scheme, kahan_scheme) else ""}
                                
                                // Load xi from device global memory.
                                // Remember that we use an interleaved memory scheme where
                                // xi = [ x1i, x2i, x3i, ... ].                  
                            	{varloader.load_vars('i', xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                            }}
                            
                            // Step 2 : Load in Shared Memory the information needed in the current block of the product -----------
                            // In the 1D scheme, we use a loop to run through the line.
                            // In the 2D scheme presented here, the computation is done in parallel wrt both lines and columns.
                            // Hence, we use "blockId.y" to get our current column number.
                            int j = blockIdx.y * blockDim.x + threadIdx.x; // Same blockDim in x and y : squared tiles.
                            if(j<ny) {{ // we load yj from device global memory only if j<ny
                                {varloader.load_vars("j", yjloc, args, row_index=j)} // load yj variables from global memory to shared memory
                            }}
                            // More precisely : the j-th line of py is loaded to yj, at a location which depends on the
                            // current threadId.
                            
                            __syncthreads(); // Make sure nobody lags behind
                            
                            // Step 3 : Once the data is loaded, execute fun --------------------------------------------------------
                            // N.B.: There's no explicit summation here. Just calls to fun, which *accumulates* the results
                            //       along the line, but does not *have* to use a "+=" as reduction operator.
                            //       In the future, we could provide other reductions: max, min, ... whatever's needed.
                            if(i<nx) {{ // we compute x1i only if needed
                                {dtype}* yjrel = yj; // Loop on the columns of the current block.
                                for(int jrel = 0; (jrel<blockDim.x) && ((blockDim.x*blockIdx.y+jrel)< ny); jrel++, yjrel+={dimy}) {{
                                    {red_formula.formula(fout,table)} // Call the function, which outputs results in fout
                                    {sum_scheme.accumulate_result(acc, fout, jrelloc, hack=True)}
                                }}
                            }}
                            __syncthreads();

                            // Step 4 : Save the result in global memory -----------------------------------------------------------
                            // The current thread has computed the "linewise-sum" of a small block of the full Kernel Product
                            // matrix, which corresponds to KP[ blockIdx.x * blockDim.x : (blockIdx.x+1) * blockDim.x ,
                            //                                  blockIdx.y * blockDim.x : (blockIdx.y+1) * blockDim.x ]
                            // We accumulate it in the output array out, which has in fact gridSize.y * nx
                            // lines of size DIMRED. The final reduction, which "sums over the block lines",
                            // shall be done in a later step.
                            if(i<nx) {{
                                {use_pragma_unroll()}
                                for(int k=0; k<{dimred}; k++) {{
                                    out[blockIdx.y*{dimred}*nx+i*{dimred}+k] = acc[k];
                                }}
                            }}
                        }}
                    """



from keopscore.utils.code_gen_utils import new_c_varname, c_array
from keopscore.utils.Tree import Tree
from keopscore import debug_ops, debug_ops_at_exec
from keopscore.utils.misc_utils import KeOps_Error

###################
## Base class
###################


class Operation(Tree):
    """Base class for all keops building block operations in a formula"""

    def __init__(self, *args, params=()):
        # *args are other instances of Operation, they are the child operations of self
        self.children = args
        self.params = params

        # The variables in the current formula is the union of the variables in the child operations.
        # Note that this requires implementing properly __eq__ and __hash__ methods in Var class.
        # N.B. We need to sort according to ind.
        set_vars = (
            set.union(*(set(arg.Vars_) for arg in args)) if len(args) > 0 else set()
        )
        self.Vars_ = sorted(list(set_vars), key=lambda v: v.ind)

    def Vars(self, cat="all"):
        # if cat=="all", returns the list of all variables in a formula, stored in self.Vars_
        # if cat is an integer between 0 and 2, returns the list of variables v such that v.cat=cat
        if cat == "all":
            return self.Vars_
        else:
            res = []
            for v in self.Vars_:
                if v.cat == cat:
                    res.append(v)
            return res

    def replace(self, old, new):
        # replace all occurences of subformula old by new in self.
        if self == old:
            return new
        else:
            new_children = [child.replace(old, new) for child in self.children]
            return type(self)(*new_children, *self.params)

    def __call__(self, out, table):
        """returns the C++ code string corresponding to the evaluation of the formula
         - out is a c_variable in which the result of the evaluation is stored
         - table is the list of c_variables corresponding to actual local variables
        required for evaluation : each Var(ind,*,*) corresponds to table[ind]"""
        from keopscore.formulas.variables.Var import Var

        string = f"\n{{\n// Starting code block for {self.__repr__()}.\n\n"
        if debug_ops:
            print(f"Building code block for {self.__repr__()}")
            print("out=", out)
            print("dim of out : ", out.dim)
            print("table=", table)
            for v in table:
                print(f"dim of {v} : ", v.dim)
        if debug_ops_at_exec:
            string += f'printf("\\n\\nComputing {self.__repr__()} :\\n");\n'
        args = []
        # Evaluation of the child operations
        for child in self.children:
            if isinstance(child, Var):
                # if the child of the operation is a Var, we do not need to evaluate it,
                # we simply record the corresponding c_variable
                arg = table[child.ind]
            else:
                # otherwise, we need to evaluate the child operation.
                # We first create a new c_array to store the result of the child operation.
                # This c_array must have a unique name in the code, to avoid conflicts
                # when we will recursively evaluate nested operations.
                template_string_id = "out_" + child.string_id.lower()
                arg_name = new_c_varname(template_string_id)
                arg = c_array(out.dtype, child.dim, arg_name)
                # Now we append into string the C++ code to declare the array
                string += f"{arg.declare()}\n"
                # Now we evaluate the child operation and append the result into string
                string += child(arg, table)
            args.append(arg)
        # Finally, evaluation of the operation itself
        string += self.Op(out, table, *args)

        # some debugging helper :
        if debug_ops_at_exec:
            for arg in args:
                string += arg.c_print
            string += out.c_print
            string += f'printf("\\n\\n");\n'
        if debug_ops:
            print(f"Finished building code block for {self.__repr__()}")

        string += f"\n\n// Finished code block for {self.__repr__()}.\n}}\n\n"
        return string

    def __mul__(self, other):
        """f*g redirects to Mult(f,g)"""
        from keopscore.formulas.maths.Mult import Mult

        return Mult(self, int2Op(other))

    def __rmul__(self, other):
        from keopscore.formulas.maths.Mult import Mult

        return Mult(int2Op(other), self)

    def __truediv__(self, other):
        """f/g redirects to Divide(f,g)"""
        from keopscore.formulas.maths.Divide import Divide

        return Divide(self, int2Op(other))

    def __rtruediv__(self, other):
        if other == 1:
            from keopscore.formulas.maths.Inv import Inv

            return Inv(self)
        else:
            return int2Op(other) / self

    def __add__(self, other):
        """f+g redirects to Add(f,g)"""
        from keopscore.formulas.maths.Add import Add

        return Add(self, int2Op(other))

    def __radd__(self, other):
        """f+g redirects to Add(f,g)"""
        return int2Op(other) + self

    def __sub__(self, other):
        """f-g redirects to Subtract(f,g)"""
        from keopscore.formulas.maths.Subtract import Subtract

        return Subtract(self, int2Op(other))

    def __rsub__(self, other):
        """f-g redirects to Subtract(f,g)"""
        return int2Op(other) - self

    def __neg__(self):
        """-f redirects to Minus(f)"""
        from keopscore.formulas.maths.Minus import Minus

        return Minus(self)

    def __pow__(self, other):
        if other == 2:
            """f**2 redirects to Square(f)"""
            from keopscore.formulas.maths.Square import Square

            return Square(self)
        elif isinstance(other, int):
            """f**m with m integer redirects to Pow(f,m)"""
            from keopscore.formulas.maths.Pow import Pow

            return Pow(self, other)
        else:
            from keopscore.formulas.maths.Powf import Powf

            raise Powf(self, other)

    def __or__(self, other):
        """f|g redirects to Scalprod(f,g)"""
        from keopscore.formulas.maths.Scalprod import Scalprod

        return Scalprod(self, other)

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.children == other.children
            and self.params == other.params
        )

    def Op(self, out, table, param):
        pass

    def chunked_version(self, dimchk):
        return None

    @property
    def is_chunkable(self):
        return False

    def chunked_formulas(self, dimchk):
        res = []
        for child in self.children:
            res += child.chunked_formulas(dimchk)
        return res

    @property
    def num_chunked_formulas(self):
        return sum([child.num_chunked_formulas for child in self.children])

    def post_chunk_formula(self, ind):
        args = []
        for child in self.children:
            args.append(child.post_chunk_formula(ind))
            ind += child.num_chunked_formulas
        return type(self)(*args, *self.params)

    enable_test = False


def int2Op(x):
    if isinstance(x, int):
        from keopscore.formulas.variables.IntCst import IntCst

        return IntCst(x)
    elif isinstance(x, Operation):
        return x
    else:
        KeOps_Error("invalid type : " + str(type(x)))


##########################
#####    Broadcast    ####
##########################

# N.B. this is used internally
def Broadcast(arg, dim):
    from keopscore.formulas.maths import SumT

    if arg.dim == dim or dim == 1:
        return arg
    elif arg.dim == 1:
        return SumT(arg, dim)
    else:
        KeOps_Error("dimensions are not compatible for Broadcast operation")



from keopscore.utils.code_gen_utils import infinity, c_if
from keopscore.formulas.reductions.Reduction import Reduction
from keopscore.utils.misc_utils import KeOps_Error


class Min_Reduction(Reduction):
    """Implements the min reduction operation : for each i or each j, find the minimal value of Fij
    operation is vectorized: if Fij is vector-valued, min is computed for each dimension."""

    string_id = "Min_Reduction"

    def __init__(self, formula, tagIJ):
        super().__init__(formula, tagIJ)
        self.dim = formula.dim  # dimension of final output of reduction
        self.dimred = self.dim  # dimension of inner reduction variables

    def InitializeReduction(self, acc):
        # Returns C++ code to be used at initialization phase of the reduction.
        # Here it consists in setting the output array to -infinity.
        return acc.assign(infinity(acc.dtype))

    def ReducePairScalar(self, acc, xi):
        # Subroutine of ReducePairShort and ReducePair methods.
        if xi.dtype == "half2":
            KeOps_Error("not implemented")
        return c_if(xi < acc, acc.assign(xi))



from keopscore.formulas.variables import Zero
from keopscore.formulas.reductions.Reduction import Reduction


class Zero_Reduction(Reduction):
    """Implements the zero reduction operation (fills output with zeros).
    N.B. The actual code for filling zeros is not here ; when a Zero_reduction is detected,
    the map_reduce scheme is redirected to CpuAssignZero or GpuAssignZero"""

    string_id = "Zero_Reduction"

    def __init__(self, dim, tagIJ):
        super().__init__(Zero(dim), tagIJ)
        self.dim = dim

    def DiffT(self, v, gradin, f0=None):
        return Zero_Reduction(v.dim, v.cat % 2)



from keopscore.binders.nvrtc.Gpu_link_compile import Gpu_link_compile
from keopscore.mapreduce.gpu.GpuAssignZero import GpuAssignZero
from keopscore.mapreduce.MapReduce import MapReduce
from keopscore.utils.code_gen_utils import (
    c_variable,
    c_array,
    c_include,
)


class GpuReduc1D_ranges(MapReduce, Gpu_link_compile):
    # class for generating the final C++ code, Gpu version

    AssignZero = GpuAssignZero

    def __init__(self, *args):
        MapReduce.__init__(self, *args)
        Gpu_link_compile.__init__(self)
        self.dimy = self.varloader.dimy

    def get_code(self):

        super().get_code()

        red_formula = self.red_formula
        dtype = self.dtype
        varloader = self.varloader

        i = self.i
        j = self.j
        fout = self.fout
        outi = self.outi
        acc = self.acc
        arg = self.arg
        args = self.args
        sum_scheme = self.sum_scheme

        nvarsi, nvarsj, nvarsp = (
            len(self.varloader.Varsi),
            len(self.varloader.Varsj),
            len(self.varloader.Varsp),
        )
        nvars = nvarsi + nvarsj + nvarsp

        param_loc = self.param_loc
        xi = self.xi
        yjloc = c_array(dtype, varloader.dimy, f"(yj + threadIdx.x * {varloader.dimy})")
        yjrel = c_array(dtype, varloader.dimy, "yjrel")
        table = varloader.table(self.xi, yjrel, self.param_loc)
        jreltile = c_variable("int", "(jrel + tile * blockDim.x)")

        indices_i = c_array("int", nvarsi, "indices_i")
        indices_j = c_array("int", nvarsj, "indices_j")
        indices_p = c_array("int", nvarsp, "indices_p")

        declare_assign_indices_i = "int *indices_i = offsets;" if nvarsi > 0 else ""
        declare_assign_indices_j = (
            f"int *indices_j = offsets + {nvarsi};" if nvarsj > 0 else ""
        )
        declare_assign_indices_p = (
            f"int *indices_p = offsets + {nvarsi} + {nvarsj};" if nvarsp > 0 else ""
        )

        starty = c_variable("int", "start_y")

        threadIdx_x = c_variable("int", "threadIdx.x")

        if dtype == "half2":
            self.headers += c_include("cuda_fp16.h")

        self.code = f"""
                        {self.headers}

                        extern "C" __global__ void GpuConv1DOnDevice_ranges(int nx, int ny, int nbatchdims,
                                                    int *offsets_d, int *lookup_d, int *slices_x,
                                                    int *ranges_y, {dtype} *out, {dtype} **{arg.id}) {{
                                                        
                          int offsets[{nvars}];
                          {declare_assign_indices_i}
                          {declare_assign_indices_j}
                          {declare_assign_indices_p}
                          
                          if (nbatchdims > 0)
                              for (int k = 0; k < {nvars}; k++)
                                  offsets[k] = offsets_d[ {nvars} * blockIdx.x + k ];
                          // Retrieve our position along the laaaaarge [1,~nx] axis: -----------------
                          int range_id= (lookup_d)[3*blockIdx.x] ;
                          int start_x = (lookup_d)[3*blockIdx.x+1] ;
                          int end_x   = (lookup_d)[3*blockIdx.x+2] ;
  
                          // The "slices_x" vector encodes a set of cutting points in
                          // the "ranges_y" array of ranges.
                          // As discussed in the Genred docstring, the first "0" is implicit:
                          int start_slice = range_id < 1 ? 0 : slices_x[range_id-1];
                          int end_slice   = slices_x[range_id];

                          // get the index of the current thread
                          int i = start_x + threadIdx.x;

                          // declare shared mem
                          extern __shared__ {dtype} yj[];
                          
                          // load parameter(s)
                          {param_loc.declare()}

                          if (nbatchdims == 0) {{
                              {varloader.load_vars("p", param_loc, args)}
                          }} else {{
                              {varloader.load_vars("p", param_loc, args, offsets=indices_p)}
                          }}
                          {fout.declare()}
                          {xi.declare()}
                          {acc.declare()}
                          {sum_scheme.declare_temporary_accumulator()}

                          if(i<end_x) {{
                              {red_formula.InitializeReduction(acc)} // acc = 0
                              {sum_scheme.initialize_temporary_accumulator_first_init()}
                              if (nbatchdims == 0) {{
                                  {varloader.load_vars('i', xi, args, row_index=i)} // load xi variables from global memory to local thread memory
                              }} else {{
                                  {varloader.load_vars('i', xi, args, row_index=threadIdx_x, offsets=indices_i)}  // Possibly, with offsets as we support broadcasting over batch dimensions
                              }}
                          }}
                          
                          int start_y = ranges_y[2*start_slice], end_y = 0;
                          for( int index = start_slice ; index < end_slice ; index++ ) {{
                              if( (index+1 >= end_slice) || (ranges_y[2*index+2] != ranges_y[2*index+1]) ) {{
                                  //start_y = ranges_y[2*index] ;
                                  end_y = ranges_y[2*index+1];

                                  for(int jstart = start_y, tile = 0; jstart < end_y; jstart += blockDim.x, tile++) {{
                                      // get the current column
                                      int j = jstart + threadIdx.x;

                                      if(j<end_y) // we load yj from device global memory only if j<end_y
                                          if (nbatchdims == 0) {{
                                              {varloader.load_vars('j', yjloc, args, row_index=j)} // load yj variables from global memory to shared memory
                                          }} else {{
                                              {varloader.load_vars('j', yjloc, args, row_index=j-starty, offsets=indices_j)}  // Possibly, with offsets as we support broadcasting over batch dimensions
                                          }}
                                      __syncthreads();
                                      
                                      if(i<end_x) {{ // we compute x1i only if needed
                                          {dtype} * yjrel = yj; // Loop on the columns of the current block.
                                          {sum_scheme.initialize_temporary_accumulator_block_init()}
                                          if (nbatchdims == 0) {{
                                              for(int jrel = 0; (jrel < blockDim.x) && (jrel<end_y-jstart); jrel++, yjrel+={varloader.dimy}) {{
                                                  {red_formula.formula(fout,table)} // Call the function, which outputs results in xi[0:DIMX1]
                                                  {sum_scheme.accumulate_result(acc, fout, jreltile+starty)}
                                              }} 
                                          }} else {{
                                              for(int jrel = 0; (jrel < blockDim.x) && (jrel<end_y-jstart); jrel++, yjrel+={varloader.dimy}) {{
                                                  {red_formula.formula(fout,table)} // Call the function, which outputs results in fout
                                                  {sum_scheme.accumulate_result(acc, fout, jreltile)}
                                              }}
                                          }}
                                          {sum_scheme.final_operation(acc)}
                                      }}
                                      __syncthreads();
                                  }}
                                  if(index+1 < end_slice)
                                      start_y = ranges_y[2*index+2] ;
                              }}
                          }}
                          if(i<end_x) {{
                          	{red_formula.FinalizeOutput(acc, outi, i)} 
                          }}
                      }}
                    """



class Tree:
    """a custom class for handling a tree structure.
    Currently we use it only to recursively print a formula or reduction"""

    def recursive_str(self):
        if hasattr(self, "print_spec"):
            idstr, mode, level = self.print_spec
            if mode == "pre":
                pre_string = idstr
                middle_string = ","
                post_string = ""
            elif mode == "mid":
                pre_string = ""
                middle_string = idstr
                post_string = ""
            elif mode == "post":
                pre_string = ""
                middle_string = ","
                post_string = idstr
        else:
            pre_string = self.string_id + "("
            middle_string = ","
            post_string = ")"
        string = pre_string
        for k, child in enumerate(self.children):
            test = (
                hasattr(child, "print_spec")
                and hasattr(self, "print_spec")
                and child.print_spec[2] >= level
            )
            string += "(" if test else ""
            string += child.recursive_str()
            string += ")" if test else ""
            string += middle_string if k < len(self.children) - 1 else ""
        for k, param in enumerate(self.params):
            if k > 0 or len(self.children) > 0:
                string += ","
            string += str(param)
        string += post_string
        return string

    def print_expand(self, depth=0):
        depth += 1
        string = self.string_id
        for child in self.children:
            string += (
                "\n" + depth * 4 * " " + "{}".format(child.recursive_str(depth=depth))
            )
        for param in self.params:
            string += "\n" + depth * 4 * " " + str(param)
        return string

    def __str__(self):
        return self.recursive_str()

    def __repr__(self):
        return self.recursive_str()

    # custom __eq__ method
    def __eq__(self, other):
        return (
            type(self) == type(other)
            and len(self.children) == len(other.children)
            and all([x == y for x, y in zip(self.children, other.children)])
        )



from keopscore.utils.code_gen_utils import (
    c_for_loop,
    new_c_varname,
    c_variable,
)

import keopscore.config.config


def math_function(
    cpu_code, gpu_code=None, gpu_half2_code=None, gpu_float_code=None, void=False
):
    if gpu_code is None:
        gpu_code = cpu_code
    if gpu_half2_code is None:
        gpu_half2_code = gpu_code
    if gpu_float_code is None:
        gpu_float_code = gpu_code

    def convert_to_fun(code):
        if isinstance(code, str):
            code_fun = lambda *args: code + "(" + ",".join(arg for arg in args) + ")"
        else:
            code_fun = code
        return code_fun

    def call(*args):
        args = list(args)
        for k, arg in enumerate(args):
            if isinstance(arg, int):
                args[k] = c_variable("int", str(arg))
        # N.B. first argument gives main dtype
        dtype = args[0].dtype
        if dtype == "half2":
            code_fun_gpu = convert_to_fun(gpu_half2_code)
        elif dtype == "float":
            code_fun_gpu = convert_to_fun(gpu_float_code)
        else:
            code_fun_gpu = convert_to_fun(gpu_code)
        string_gpu = code_fun_gpu(*(arg.id for arg in args))
        code_fun_cpu = convert_to_fun(cpu_code)
        string_cpu = code_fun_cpu(*(arg.id for arg in args))
        string = f"""
                    #ifdef __CUDACC__
                        {string_gpu}
                    #else
                        {string_cpu}
                    #endif
                """
        if void:
            return string
        else:
            return c_variable(dtype, string)

    return call


keops_mul = math_function(cpu_code=lambda x, y: f"({x}*{y})", gpu_half2_code="__hmul2")

keops_abs = math_function(cpu_code="abs", gpu_half2_code="__habs2")
keops_cos = math_function(cpu_code="cos", gpu_half2_code="h2cos")
keops_sin = math_function(cpu_code="sin", gpu_half2_code="h2sin")
keops_sinxdivx = math_function(cpu_code=lambda x: f"({x} ? sin({x})/{x} : 1.0f)")
keops_acos = math_function(cpu_code="acos")
keops_asin = math_function(cpu_code="asin")
keops_atan = math_function(cpu_code="atan")
keops_atan2 = math_function(cpu_code="atan2")
keops_exp = math_function(cpu_code="exp", gpu_half2_code="h2exp")
keops_floor = math_function(cpu_code="floor")
keops_log = math_function(cpu_code="log")
keops_xlogx = math_function(cpu_code=lambda x: f"({x} ? {x} * log({x}) : 0.0f)")
keops_fma = math_function(cpu_code="fma")
keops_pow = math_function(
    cpu_code="pow",
    gpu_code="powf",
    gpu_half2_code=lambda x, y: f"h2exp(__float2half2_rn((float){y})*h2log({x}))",
)
keops_powf = math_function(
    cpu_code="powf", gpu_half2_code=lambda x, y: f"h2exp({y}*h2log({x}))"
)
keops_rcp = math_function(cpu_code=lambda x: f"(1.0f/({x}))", gpu_half2_code="h2rcp")

keops_rsqrt = math_function(
    cpu_code=lambda x: f"(({x}==0.0f)? 0.0f : 1.0f/sqrt({x}))",
    gpu_code=lambda x: f"(({x}==0.0f)? 0.0f : rsqrt({x}))",
    gpu_half2_code=lambda x: f"h2rsqrt({x}+__heq2({x},__float2half2_rn(0.0f))) * (__float2half2_rn(1.0f)-__heq2({x},__float2half2_rn(0.0f)))",
)

keops_sqrt = math_function(cpu_code="sqrt", gpu_half2_code="h2sqrt")

keops_relu = math_function(cpu_code=lambda x: f"(({x}<0.0f)? 0.0f : {x})")
keops_step = math_function(cpu_code=lambda x: f"(({x}<0.0f)? 0.0f : 1.0f)")
keops_sign = math_function(
    cpu_code=lambda x: f"(({x}>0.0f)? 1.0f : ( ({x}<0.0f)? -1.0f : 0.0f ))"
)
keops_clamp = math_function(
    cpu_code=lambda x, a, b: f"(({x}<{a})? {a} : ( ({x}>{b})? {b} : {x} ))"
)
keops_clampint = math_function(
    cpu_code=lambda x, a, b: f"(({x}<{a})? {a} : ( ({x}>{b})? {b} : {x} ))"
)
keops_mod = math_function(
    cpu_code=lambda x, n, d: f"({x} - {n} * floor(({x} - {d})/{n}))"
)
keops_round = math_function(
    cpu_code=lambda x, d: f"round({x})"
    if eval(d) == 0
    else f"(round({x}*{10**eval(d)})/{10**eval(d)})"
)
keops_diffclampint = math_function(
    cpu_code=lambda x, a, b: f"(({x}<{a})? 0.0f : ( ({x}>{b})? 0.0f : 1.0f ))"
)
keops_ifelse = math_function(cpu_code=lambda x, a, b: f"(({x}>=0.0f) ? {a} : {b})")

keops_sincos = math_function(
    cpu_code=lambda x, s, c: f"*{s}=sin(x); *{c}=cos(x);", void=True
)



from .GpuAssignZero import GpuAssignZero
from .GpuReduc1D import GpuReduc1D
from .GpuReduc1D_chunks import GpuReduc1D_chunks
from .GpuReduc1D_finalchunks import GpuReduc1D_finalchunks
from .GpuReduc1D_ranges import GpuReduc1D_ranges
from .GpuReduc1D_ranges_chunks import GpuReduc1D_ranges_chunks
from .GpuReduc1D_ranges_finalchunks import (
    GpuReduc1D_ranges_finalchunks,
)
from .GpuReduc2D import GpuReduc2D
